"""The DDS response.

The DDS response builds a representation of the structure of the dataset,
informing which variables are contained, their shape, type and dimensions.
Together with the DAS, the DDS describes all metadata associated with a given
dataset, allowing clients to introspect the variables and request data as
necessary.

"""

try:
    from functools import singledispatch
except ImportError:
    from singledispatch import singledispatch

from six.moves import map, zip

from ..model import (DatasetType, BaseType,
                     SequenceType, StructureType,
                     GridType)
from .lib import BaseResponse
from ..lib import __version__, NUMPY_TO_DAP2_TYPEMAP


INDENT = ' ' * 4


class DDSResponse(BaseResponse):

    """The DDS response."""

    __version__ = __version__

    def __init__(self, dataset):
        BaseResponse.__init__(self, dataset)
        self.headers.extend([
            ('Content-description', 'dods_dds'),
            ('Content-type', 'text/plain; charset=ascii'),
        ])

    def __iter__(self):
        for line in dds(self.dataset):
            yield line.encode('ascii')


@singledispatch
def dds(var, level=0, sequence=0):
    """Single dispatcher that generates the DDS response."""
    raise StopIteration


@dds.register(DatasetType)
def _(var, level=0, sequence=0):
    yield "{indent}Dataset {{\n".format(indent=level*INDENT)
    for child in var.children():
        for line in dds(child, level+1, sequence):
            yield line
    yield "{indent}}} {name};\n".format(indent=level*INDENT, name=var.name)


@dds.register(SequenceType)
def _sequencetype(var, level=0, sequence=0):
    yield "{indent}Sequence {{\n".format(indent=level*INDENT)
    for child in var.children():
        for line in dds(child, level+1, sequence+1):
            yield line
    yield "{indent}}} {name};\n".format(indent=level*INDENT, name=var.name)


@dds.register(StructureType)
def _structuretype(var, level=0, sequence=0):
    yield "{indent}Structure {{\n".format(indent=level*INDENT)
    for child in var.children():
        for line in dds(child, level+1, sequence):
            yield line
    yield "{indent}}} {name};\n".format(indent=level*INDENT, name=var.name)


@dds.register(GridType)
def _gridtype(var, level=0, sequence=0):
    yield '{indent}Grid {{\n'.format(indent=level*INDENT)

    yield '{indent}Array:\n'.format(indent=(level+1)*INDENT)
    for line in dds(var.array, level+2, sequence):
        yield line

    yield '{indent}Maps:\n'.format(indent=(level+1)*INDENT)
    for map_ in var.maps.values():
        for line in dds(map_, level+2, sequence):
            yield line

    yield '{indent}}} {name};\n'.format(indent=level*INDENT, name=var.name)


@dds.register(BaseType)
def _basetype(var, level=0, sequence=0):
    shape = var.shape[sequence:]

    if var.dimensions:
        shape = ''.join(
            map('[{0[0]} = {0[1]}]'.format, zip(var.dimensions, shape)))
    elif len(shape) == 1:
        shape = '[{0} = {1}]'.format(var.name, shape[0])
    else:
        shape = ''.join('[{0}]'.format(len) for len in shape)

    yield '{indent}{type} {name}{shape};\n'.format(
        indent=level*INDENT,
        type=NUMPY_TO_DAP2_TYPEMAP[var.dtype.char],
        name=var.name,
        shape=shape)



"""Test the DAP handler, which forms the core of the client."""

import numpy as np
from six.moves import zip
from pydap.handlers.dap import DAPHandler
import pytest


@pytest.fixture(scope='module')
def simple_data():
    data = [(10, 15.2, 'Diamond_St'),
            (11, 13.1, 'Blacktail_Loop'),
            (12, 13.3, 'Platinum_St'),
            (13, 12.1, 'Kodiak_Trail')]
    return data


@pytest.fixture(scope='module')
def simple_nc_file(simple_data, tmpdir_factory):
    from netCDF4 import Dataset
    file_name = str(tmpdir_factory.mktemp('nc').join('simple.nc'))
    with Dataset(file_name, 'w') as output:
        output.createDimension('index', None)
        temp = output.createVariable('index', '<i4', ('index',))
        split_data = zip(*simple_data)
        temp[:] = next(split_data)
        temp = output.createVariable('temperature', '<f8', ('index',))
        temp[:] = next(split_data)
        temp = output.createVariable('station', 'S40', ('index',))
        for item_id, item in enumerate(next(split_data)):
            temp[item_id] = item
    return file_name


@pytest.fixture(scope='module')
def simple_handler(simple_nc_file):
    from pydap.handlers.netcdf import NetCDFHandler
    return NetCDFHandler(simple_nc_file)


def test_handler(simple_data, simple_handler):
    """Test that dataset has the correct data proxies for grids."""
    dataset = simple_handler.dataset
    dtype = [('index', '<i4'),
             ('temperature', '<f8'),
             ('station', 'S40')]
    retrieved_data = list(zip(dataset['index'][:],
                              dataset['temperature'].array[:],
                              dataset['station'].array[:]))
    np.testing.assert_array_equal(np.array(retrieved_data, dtype=dtype),
                                  np.array(simple_data, dtype=dtype))


@pytest.fixture(scope='module')
def simple_application(simple_handler):
    from pydap.wsgi.ssf import ServerSideFunctions
    return ServerSideFunctions(simple_handler)


def test_open(simple_data, simple_application):
    """Test that NetCDFHandler can be read through open_url."""
    dataset = DAPHandler("http://localhost:8001/", simple_application).dataset
    dtype = [('index', '<i4'),
             ('temperature', '<f8'),
             ('station', 'S40')]
    retrieved_data = list(zip(dataset['index'][:],
                              dataset['temperature'].array[:],
                              dataset['station'].array[:]))
    np.testing.assert_array_equal(np.array(retrieved_data, dtype=dtype),
                                  np.array(simple_data, dtype=dtype))



import unittest

import numpy as np
from pydap.parsers import parse_ce
from pydap.handlers.lib import BaseHandler
from pydap.tests.datasets import D1


class TestConstrain(unittest.TestCase):
    def test_no_ce(self):
        data = np.rec.fromrecords(
                D1.Drifters.data.tolist(), names=list(D1.Drifters.keys()))

        projection, selection = parse_ce('')
        dataset = BaseHandler(D1).parse(projection, selection)
        np.testing.assert_array_equal(data, dataset.Drifters.data)

    def test_filtering(self):
        data = np.rec.fromrecords(
                D1.Drifters.data.tolist(), names=list(D1.Drifters.keys()))
        filtered = data[data['longitude'] < 999]

        projection, selection = parse_ce('Drifters.longitude<999')
        dataset = BaseHandler(D1).parse(projection, selection)
        np.testing.assert_array_equal(filtered, dataset.Drifters.data)



from webob.request import Request
from webob.exc import HTTPError
import threading
import multiprocessing
import time
import math
import numpy as np
import socket

from wsgiref.simple_server import make_server

from ..handlers.lib import BaseHandler
from ..model import BaseType, DatasetType
from ..net import get_response

DefaultDataset = DatasetType("Default")
DefaultDataset["byte"] = BaseType("byte", np.arange(5, dtype="B"))
DefaultDataset["string"] = BaseType("string", np.array(["one", "two"]))
DefaultDataset["short"] = BaseType("short", np.array(1, dtype="h"))


def get_open_port():
    # http://stackoverflow.com/questions/2838244/
    # get-open-tcp-port-in-python/2838309#2838309
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port


def run_server_in_process(httpd, shutdown, **kwargs):
    _server = (threading
               .Thread(target=httpd.serve_forever,
                       kwargs=kwargs))
    _server.start()
    shutdown.wait()
    httpd.shutdown()
    _server.join()


class LocalTestServer(object):
    """
    Simple server instance that can be used to test pydap.
    Relies on threading and is usually slow (it has to
    start and shutdown which typically takes ~2 sec).

    Usage:
    >>> import numpy as np
    >>> from pydap.handlers.lib import BaseHandler
    >>> from pydap.model import DatasetType, BaseType
    >>> DefaultDataset = DatasetType("Default")
    >>> DefaultDataset["byte"] = BaseType("byte", np.arange(5, dtype="B"))
    >>> DefaultDataset["string"] = BaseType("string", np.array(["one", "two"]))
    >>> DefaultDataset["short"] = BaseType("short", np.array(1, dtype="h"))
    >>> DefaultDataset
    <DatasetType with children 'byte', 'string', 'short'>
    >>> application = BaseHandler(DefaultDataset)
    >>> from pydap.client import open_url

    As an instance:
    >>> with LocalTestServer(application) as server:
    ...     dataset = open_url("http://localhost:%s" % server.port)
    ...     dataset
    ...     print(dataset['byte'].data[:])
    ...     print(dataset['string'].data[:])
    ...     print(dataset['short'].data[:])
    <DatasetType with children 'byte', 'string', 'short'>
    [0 1 2 3 4]
    [b'one' b'two']
    1

    Or by managing connection and deconnection:
    >>> server = LocalTestServer(application)
    >>> server.start()
    >>> dataset = open_url("http://localhost:%s" % server.port)
    >>> dataset
    <DatasetType with children 'byte', 'string', 'short'>
    >>> print(dataset['byte'].data[:])
    [0 1 2 3 4]
    >>> server.shutdown()
    """

    def __init__(self, application=BaseHandler(DefaultDataset),
                 port=None, wait=0.5, polling=1e-2, as_process=False):
        self._port = port or get_open_port()
        self.application = application
        self._wait = wait
        self._polling = polling
        self._as_process = as_process
        self._address = '0.0.0.0'

    @property
    def url(self):
        return "http://{0}:{1}/".format(self._address, self.port)

    def start(self):
        # Start a simple WSGI server:
        application = self.application

        self._httpd = make_server(self._address, self.port, application)
        kwargs = {'poll_interval': 0.1}

        if self._as_process:
            self._shutdown = multiprocessing.Event()
            self._server = (multiprocessing
                            .Process(target=run_server_in_process,
                                     args=(self._httpd, self._shutdown),
                                     kwargs=kwargs))
        else:
            self._server = (threading
                            .Thread(target=self._httpd.serve_forever,
                                    kwargs=kwargs))

        self._server.start()
        self.poll_server()

    def poll_server(self):
        # Poll the server
        ok = False
        for trial in range(int(math.ceil(self._wait/self._polling))):
            try:
                # When checking whether server has started, do
                # not verify ssl:
                resp = get_response(
                        Request
                        .blank(self.url + '.dds'),
                        None, verify=False)
                ok = (resp.status_code == 200)
            except HTTPError:
                pass
            if ok:
                break
            time.sleep(self._polling)

        if not ok:
            self.shutdown()
            raise Exception(('LocalTestServer did not start in {0}s. '
                             'Try using LocalTestServer(..., wait={1}')
                            .format(self._wait, 2*self._wait))

    @property
    def port(self):
        return self._port

    def __enter__(self):
        self.start()
        return self

    def shutdown(self):
        # Tell the server to shutdown:
        if self._as_process:
            self._shutdown.set()
        else:
            self._httpd.shutdown()
        self._server.join()
        self._httpd.server_close()

    def __exit__(self, *_):
        self.shutdown()



"""Response with information about pydap version."""

import sys
from six import text_type

from webob import Response
from json import dumps
from pkg_resources import iter_entry_points

from ..lib import __version__, __dap__


class VersionResponse(object):

    """A specialized response for debugging.

    This is a special response used to display information about the server.

    """

    __version__ = __version__

    def __init__(self, dataset):
        output = {
            "pydap": __version__,
            "dap": __dap__,
            "handlers": dict(
                (ep.name, getattr(ep.load(), "__version__", "Unknown"))
                for ep in iter_entry_points("pydap.handler")
            ),
            "responses": dict(
                (ep.name, getattr(ep.load(), "__version__", "Unknown"))
                for ep in iter_entry_points("pydap.response")
            ),
            "functions": dict(
                (ep.name, getattr(ep.load(), "__version__", "Unknown"))
                for ep in iter_entry_points("pydap.function")
            ),
            "python": sys.version,
        }
        self.body = dumps(output, indent=4)

    def __call__(self, environ, start_response):
        res = Response()
        res.text = text_type(self.body)
        res.status = '200 OK'
        res.content_type = 'application/json'
        res.charset = 'utf-8'
        res.headers.add('Content-description', 'dods_version')
        res.headers.add('XDODS-Server', 'pydap/%s' % __version__)

        return res(environ, start_response)



from . import get_cookies


def setup_session(username, password, check_url=None,
                  session=None, verify=True):
    """
    A special call to get_cookies.setup_session that is tailored for
    URS EARTHDATA at NASA credentials.

    """
    if session is not None:
        # URS connections cannot be kept alive at the moment.
        session.headers.update({'Connection': 'close'})
    session = get_cookies.setup_session('https://urs.earthdata.nasa.gov',
                                        username=username,
                                        password=password,
                                        session=session,
                                        check_url=check_url,
                                        verify=verify)
    return session



import pickle
import codecs


def main():
    logfile = 'dap4_access_test.log'

    with open(logfile, 'r') as f:
        f = f.read()

    p = f.split('----BEGIN PICKLE-----')[1].split('----END PICKLE-----')[0]
    p = ''.join(p).strip().replace('\n', '')

    response = pickle.loads(codecs.decode(p.encode('utf-8'), "base64"))

    print(dict(response.headers))
    with open('out.html', 'w') as f:
        f.write(response.text)

    with open('response.pickle', 'wb') as f:
        pickle.dump(response, f)


if __name__ == '__main__':
    main()



"""Test basic handler functions."""

import copy
from six import text_type

from webtest import AppError
from webtest import TestApp as App
import numpy as np

from pydap.model import BaseType, StructureType, SequenceType
from pydap.lib import walk
from pydap.exceptions import ConstraintExpressionError
from pydap.handlers.lib import (
    load_handlers, get_handler, BaseHandler, ExtensionNotSupportedError,
    apply_selection, apply_projection, ConstraintExpression,
    IterData)
from pydap.parsers import parse_projection
from pydap.tests.datasets import (
    SimpleArray, SimpleSequence, SimpleGrid, VerySimpleSequence,
    NestedSequence, SimpleStructure)
import unittest


class TestHandlersLib(unittest.TestCase):

    """Test handler loading."""

    def test_load_handlers(self):
        """Test that handlers can be loaded correctly.

        We use a mock working set, since by default no handlers are installed
        with pydap.

        """
        handlers = load_handlers(MockWorkingSet())
        self.assertTrue(MockHandler in handlers)

    def test_get_handler(self):
        """Test that we can load a specific handler."""
        handlers = load_handlers(MockWorkingSet())
        handler = get_handler("file.foo", handlers)
        self.assertIsInstance(handler, MockHandler)

    def test_no_handler_available(self):
        """Test exception raised when file not supported."""
        with self.assertRaises(ExtensionNotSupportedError):
            get_handler("file.bar")


class TestBaseHandler(unittest.TestCase):

    """Test the base handler as a WSGI app."""

    def setUp(self):
        """Create a basic WSGI app."""
        self.app = App(MockHandler(SimpleArray))

    def test_unconstrained_das(self):
        """DAS responses are always unconstrained."""
        res = self.app.get("/.dds")
        self.assertEqual(res.text, """Dataset {
    Byte byte[byte = 5];
    String string[string = 2];
    Int16 short;
} SimpleArray;
""")

        res = self.app.get("/.dds?byte")
        self.assertEqual(res.text, """Dataset {
    Byte byte[byte = 5];
} SimpleArray;
""")

        res = self.app.get("/.das")
        das = res.text
        self.assertEqual(das, """Attributes {
    byte {
    }
    string {
    }
    short {
    }
}
""")

        # check that DAS is unmodifed with constraint expression
        res = self.app.get("/.das?byte")
        self.assertEqual(res.text, das)

    def test_exception(self):
        """
        Test exception handling.

        By default pydap will capture all exceptions and return a formatted
        error response.

        """
        with self.assertRaises(AppError):
            self.app.get("/.foo")

    def test_exception_non_captured(self):
        """Test exception handling when not captured."""
        app = App(MockHandler(SimpleArray), extra_environ={
            "x-wsgiorg.throw_errors": True})
        with self.assertRaises(KeyError):
            app.get("/.foo")

    def test_missing_dataset(self):
        """Test exception when dataset is not set."""
        app = App(MockHandler(), extra_environ={
            "x-wsgiorg.throw_errors": True})
        with self.assertRaises(NotImplementedError):
            app.get("/.dds")


class TestApplySelection(unittest.TestCase):

    """Test function that applies selections to the dataset."""

    def setUp(self):
        """Build our own sequence.

        pydap uses lightweight copies of objects that share data. This breaks
        unit tests since the same objects are reused for tests.

        """
        # make a dataset with its own data
        self.dataset = copy.copy(SimpleSequence)
        self.dataset.cast.data = SimpleSequence.cast.data.copy()

    def test_no_selection(self):
        """Test no selection in the query string."""
        dataset = apply_selection("", self.dataset)
        np.testing.assert_array_equal(
            dataset.cast.data, self.dataset.cast.data)

    def test_simple_selection(self):
        """Test a simple selection applied to the dataset."""
        dataset = apply_selection(["cast.lon=100"], self.dataset)
        np.testing.assert_array_equal(
            dataset.cast.data,
            self.dataset.cast.data[self.dataset.cast.data["lon"] == 100])

    def test_multiple_selections(self):
        """Test multiple selections applied to dataset."""
        dataset = apply_selection(
            ["cast.lon=100", "cast.lat>0"], self.dataset)
        np.testing.assert_array_equal(
            dataset.cast.data,
            self.dataset.cast.data[
                self.dataset.cast.data["lon"] == 100
            ][
                self.dataset.cast.data["lat"] > 0
            ])


class TestApplyProjectionGrid(unittest.TestCase):

    """Test applying projections on a dataset with a grid."""

    def setUp(self):
        """Build dataset with no shared data."""
        self.dataset = copy.copy(SimpleGrid)
        for var in walk(self.dataset, BaseType):
            var.data = var.data.copy()

    def test_no_projection(self):
        """Test no projections."""
        dataset = apply_projection("", self.dataset)
        self.assertEqual(list(dataset.children()), [])

    def test_simple_projection(self):
        """Test simple projections."""
        dataset = apply_projection(parse_projection("x"), self.dataset)
        self.assertEqual(list(dataset.keys()), ["x"])

    def test_simple_projection_with_index(self):
        """Test simple projections."""
        dataset = apply_projection(parse_projection("x[1]"), self.dataset)
        np.testing.assert_array_equal(
            dataset.x.data, [1])

    def test_array(self):
        """Test that the grid degenerates into a structure."""
        dataset = apply_projection(
            parse_projection("SimpleGrid.SimpleGrid"), self.dataset)
        self.assertIsInstance(dataset.SimpleGrid, StructureType)

    def test_array_slice(self):
        """Test slices applied to a grid."""
        dataset = apply_projection(
            parse_projection("SimpleGrid[1]"), self.dataset)
        np.testing.assert_array_equal(
            dataset.SimpleGrid.x.data, self.dataset.SimpleGrid[1].x.data)
        np.testing.assert_array_equal(
            dataset.SimpleGrid.y.data, self.dataset.SimpleGrid[1].y.data)
        np.testing.assert_array_equal(
            dataset.SimpleGrid.SimpleGrid.data,
            self.dataset.SimpleGrid[1:2].SimpleGrid.data)


class TestApplyProjectionSequence(unittest.TestCase):

    """Test applying projections on a dataset with a sequence."""

    def setUp(self):
        """Build dataset with no shared data."""
        self.dataset = copy.copy(VerySimpleSequence)
        self.dataset.sequence.data = VerySimpleSequence.sequence.data.copy()

    def test_sequence_projection(self):
        """Test projection slicing on sequences."""
        dataset = apply_projection(
            parse_projection("sequence[2]"), self.dataset)
        np.testing.assert_array_equal(
            dataset.sequence.data, VerySimpleSequence.sequence.data[2])


class TestInvalidProjection(unittest.TestCase):

    """Test applying a projection to a structure object."""

    def test_structure_projection(self):
        """Test projection slicing on a structure."""
        with self.assertRaises(ConstraintExpressionError):
            apply_projection(parse_projection("types[0]"), SimpleStructure)


class TestConstraintExpression(unittest.TestCase):

    """Test the constraint expression object."""

    def test_str(self):
        """Test string representation."""
        ce = ConstraintExpression("a>1")
        self.assertEqual(str(ce), "a>1")

    def test_unicode(self):
        """Test unicode representation."""
        ce = ConstraintExpression("a>1")
        self.assertEqual(text_type(ce), "a>1")

    def test_and(self):
        """Test CE addition."""
        ce1 = ConstraintExpression("a>1")
        ce2 = ConstraintExpression("b>0")
        ce3 = ce1 & ce2
        self.assertEqual(str(ce3), "a>1&b>0")

    def test_or(self):
        """Expressions cannot be ORed."""
        ce1 = ConstraintExpression("a>1")
        ce2 = ConstraintExpression("b>0")
        with self.assertRaises(ConstraintExpressionError):
            ce1 | ce2


class TestIterData(unittest.TestCase):
    """
    Test the ``IterData`` class, used to store flat/nested sequence data.

    A flat ``IterData`` should behave like a Numpy structured array, except
    all operations are stored to be lazily evaluated when the object is
    iterated over.

    """

    def setUp(self):
        """Create a flat IterData."""
        template = SequenceType("a")
        template["b"] = BaseType("b")
        template["c"] = BaseType("c")
        template["d"] = BaseType("d")
        self.data = IterData([(1, 2, 3), (4, 5, 6)], template)

        self.array = np.array(np.rec.fromrecords([
            (1, 2, 3),
            (4, 5, 6),
        ], names=["b", "c", "d"]))

    def assertIteratorEqual(self, it1, it2):
        self.assertEqual(list(it1), list(it2))

    def test_repr(self):
        """Test the object representation."""
        self.assertEqual(
            repr(self.data),
            "<IterData to stream [(1, 2, 3), (4, 5, 6)]>")

    def test_dtype(self):
        """Test the ``dtype`` property."""
        self.assertEqual(self.data["b"].dtype, self.array["b"].dtype)

    def test_iteration(self):
        """Test iteration over data."""
        self.assertIteratorEqual(map(tuple, self.data), map(tuple, self.array))
        self.assertIteratorEqual(self.data["b"], self.array["b"])

    def test_filter(self):
        """Test filtering the object."""
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] == 1]),
            map(tuple, self.array[self.array["b"] == 1]))
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] != 1]),
            map(tuple, self.array[self.array["b"] != 1]))
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] >= 1]),
            map(tuple, self.array[self.array["b"] >= 1]))
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] <= 1]),
            map(tuple, self.array[self.array["b"] <= 1]))
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] > 1]),
            map(tuple, self.array[self.array["b"] > 1]))
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] < 1]),
            map(tuple, self.array[self.array["b"] < 1]))

    def test_slice(self):
        """Test slicing the object."""
        self.assertIteratorEqual(map(tuple, self.data[1:]),
                                 map(tuple, self.array[1:]))

    def test_integer_slice(self):
        """Test slicing with an integer.

        Note that the behavior here is different from Numpy arrays, since the
        data access to ``IterData`` is through iteration it has no direct index
        access.

        """
        self.assertIteratorEqual(self.data[0:1], self.array[0:1].tolist())
        self.assertIteratorEqual(self.data[0], self.array[0:1].tolist())

    def test_invalid_child(self):
        """Test accessing a non-existing child."""
        with self.assertRaises(KeyError):
            self.data["e"]

    def test_invalid_key(self):
        """Test accessing using an invalid key."""
        with self.assertRaises(KeyError):
            self.data[(1, 2)]

    def test_selecting_children(self):
        """Test that we can select children."""
        self.assertIteratorEqual(
            map(tuple, self.data[["d", "b"]]),
            map(tuple, self.array[["d", "b"]]))

    def test_invalid_selection(self):
        """Test invalid selections.

        In theory this should never happen, since ``ConstraintExpression``
        object are constructly directly from existing children.

        """
        with self.assertRaises(ConstraintExpressionError):
            self.data[ConstraintExpression("a.e<1")]
        with self.assertRaises(ConstraintExpressionError):
            self.data[ConstraintExpression("a.d<foo")]

    def test_intercomparison_selection(self):
        """Test comparing children in the selection."""
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] == self.data["c"]]),
            map(tuple, self.array[self.array["b"] == self.array["c"]]))
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] != self.data["c"]]),
            map(tuple, self.array[self.array["b"] != self.array["c"]]))
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] >= self.data["c"]]),
            map(tuple, self.array[self.array["b"] >= self.array["c"]]))
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] <= self.data["c"]]),
            map(tuple, self.array[self.array["b"] <= self.array["c"]]))
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] > self.data["c"]]),
            map(tuple, self.array[self.array["b"] > self.array["c"]]))
        self.assertIteratorEqual(
            map(tuple, self.data[self.data["b"] < self.data["c"]]),
            map(tuple, self.array[self.array["b"] < self.array["c"]]))

    def test_selection_not_in_projection(self):
        """Test selection with variables that are not in the projection."""
        self.data[["d", "b"]]
        filtered = self.data[["d", "b"]][self.data["c"] > 3]
        self.assertEqual(filtered, self.array[["d", "b"]][self.array["c"] > 3])


class TestRegexp(unittest.TestCase):

    """Test regular expression match."""

    def test_regexp(self):
        sequence = SequenceType("sequence")
        sequence["name"] = BaseType("name")
        sequence.data = IterData([
            ("John", "Paul", "George", "Ringo"),
        ], sequence)

        filtered = sequence[ConstraintExpression('sequence.name=~"J.*"')]
        self.assertEqual(list(filtered.iterdata()), [("John",)])


class TestNestedIterData(unittest.TestCase):

    """Test ``IterData`` with nested data."""

    def setUp(self):
        """Load data from test dataset."""
        self.data = NestedSequence.location.data

    def test_iteration(self):
        """Test basic iteration."""
        self.assertEqual(list(self.data),
                         [(1, 1, 1, [(10, 11, 12), (21, 22, 23)]),
                          (2, 4, 4, [(15, 16, 17)]),
                          (3, 6, 9, []),
                          (4, 8, 16, [(31, 32, 33), (41, 42, 43),
                                      (51, 52, 53), (61, 62, 63)])])

    def test_children_data(self):
        """Test getting data from a simple child."""
        self.assertEqual(list(self.data["lat"]), [1, 2, 3, 4])

    def test_sequence_children_data(self):
        """Test getting data from a sequence child."""
        self.assertEqual(list(self.data["time_series"]),
                         [[(10, 11, 12), (21, 22, 23)],
                          [(15, 16, 17)],
                          [],
                          [(31, 32, 33), (41, 42, 43),
                           (51, 52, 53), (61, 62, 63)]])

    def test_deep_children_data(self):
        """Test getting data from a sequence child."""
        self.assertEqual(list(self.data["time_series"]["time"]),
                         [[10, 21], [15], [], [31, 41, 51, 61]])

    def test_selecting_children(self):
        """Test that we can select children."""
        self.assertEqual(list(self.data[["time_series", "elev"]]),
                         [([(10, 11, 12), (21, 22, 23)], 1),
                          ([(15, 16, 17)], 4),
                          ([], 9),
                          ([(31, 32, 33), (41, 42, 43),
                            (51, 52, 53), (61, 62, 63)], 16)])

    def test_slice(self):
        """Test slicing the object."""
        self.assertEqual(list(self.data[1::2]),
                         [(2, 4, 4, [(15, 16, 17)]),
                          (4, 8, 16, [(31, 32, 33), (41, 42, 43),
                                      (51, 52, 53), (61, 62, 63)])])

    def test_children_data_from_slice(self):
        """Test getting children data from a sliced sequence."""
        self.assertEqual(list(self.data[1::2]["lat"]), [2, 4])

    def test_sequence_children_data_from_slice(self):
        """Test getting children data from a sliced sequence."""
        self.assertEqual(list(self.data[1::2]["time_series"]),
                         [[(15, 16, 17)], [(31, 32, 33), (41, 42, 43),
                                           (51, 52, 53), (61, 62, 63)]])

    def test_deep_slice(self):
        """Test slicing the inner sequence."""
        self.assertEqual(list(self.data["time_series"][::2]),
                         [[(10, 11, 12), (21, 22, 23)], []])

    def test_integer_slice(self):
        """Test slicing with an integer."""
        self.assertEqual(list(self.data["time_series"][1]), [[(15, 16, 17)]])

    def test_filter_data(self):
        """Test filtering the data."""
        self.assertEqual(list(self.data[self.data["lat"] > 2]),
                         [(3, 6, 9, []), (4, 8, 16,
                                          [(31, 32, 33), (41, 42, 43),
                                           (51, 52, 53), (61, 62, 63)])])

    def test_deep_filter(self):
        """Test deep filtering the data."""
        self.assertEqual(list(self.data[self.data["time_series"]["slp"] > 11]),
                         [(1, 1, 1, [(21, 22, 23)]),
                          (2, 4, 4, [(15, 16, 17)]),
                          (3, 6, 9, []),
                          (4, 8, 16, [(31, 32, 33), (41, 42, 43),
                                      (51, 52, 53), (61, 62, 63)])])


class MockWorkingSet(object):

    """A fake working set for testing handlers."""

    def iter_entry_points(self, group):
        """Return a mock entry point."""
        yield MockEntryPoint(MockHandler)


class MockHandler(BaseHandler):

    """A fake handler for testing."""

    extensions = r"^.*\.foo$"

    def __init__(self, dataset=None):
        BaseHandler.__init__(self, dataset)
        self.additional_headers = [
            ("X-debug", "True")
        ]


class MockEntryPoint(object):

    """A fake entry point for testing."""

    def __init__(self, handler):
        self.handler = handler
        self.name = 'test'
        self.module_name = 'pydap.handlers.test'
        self.attrs = ('TestHandler', )

    def load(self):
        """Return the wrapped handler."""
        return self.handler



"""Test the DAP handler, which forms the core of the client."""

import csv
import numpy as np

from pydap.handlers.csv import CSVHandler
from pydap.handlers.dap import DAPHandler

import pytest


@pytest.fixture(scope='session')
def simple_data():
    data = [(10, 15.2, 'Diamond_St'),
            (11, 13.1, 'Blacktail_Loop'),
            (12, 13.3, 'Platinum_St'),
            (13, 12.1, 'Kodiak_Trail')]
    return data


@pytest.fixture(scope='session')
def simple_data_file(tmpdir_factory, simple_data):
    temp_file = str(tmpdir_factory.mktemp('data').join('simple_data.csv'))
    with open(temp_file, 'w') as f:
        writer = csv.writer(f, quoting=csv.QUOTE_NONNUMERIC)
        writer.writerow(['index', 'temperature', 'site'])
        for row in simple_data:
            writer.writerow(row)
    return temp_file


def test_open(simple_data, simple_data_file):
    """Test that dataset has the correct data proxies for grids."""
    dataset = DAPHandler("http://localhost:8001/",
                         CSVHandler(simple_data_file)).dataset
    seq = dataset['sequence']
    dtype = [('index', '<i4'),
             ('temperature', '<f8'),
             ('site', 'S40')]
    retrieved_data = [line for line in seq]

    np.testing.assert_array_equal(np.array(retrieved_data, dtype=dtype),
                                  np.array(simple_data, dtype=dtype))


def test_combined_slice(simple_data, simple_data_file):
    """Test that dataset has the correct data proxies for grids."""
    dataset = CSVHandler(simple_data_file).dataset
    seq = dataset['sequence']
    retrieved_data = [line for line
                      in seq[['temperature', 'site']][seq['index'] > 10]]

    dtype = [('temperature', '<f8'),
             ('site', 'S40')]

    np.testing.assert_array_equal(
                np.array(retrieved_data, dtype=dtype),
                np.array([item[1:] for item in simple_data[1:]],
                         dtype=dtype))


def test_constrained(simple_data, simple_data_file):
    """Test that dataset has the correct data proxies for grids."""
    dataset = CSVHandler(simple_data_file).dataset
    seq = dataset['sequence']
    retrieved_data = [line for line
                      in seq[seq['index'] > 10]['site'][::2]]

    dtype = 'S40'

    np.testing.assert_array_equal(
                np.array(retrieved_data, dtype=dtype),
                np.array([simple_data[idx][-1] for idx in [1, 3]],
                         dtype=dtype))



"""A DMR parser."""

import numpy as np
from xml.etree import ElementTree as ET
import pydap.model
import pydap.lib
import re
import copy
import collections

constructors = ('grid', 'sequence', 'structure')
name_regexp = r'[\w%!~"\'\*-]+'
dmr_atomic_types = ('Int8', 'UInt8', 'Byte', 'Char', 'Int16', 'UInt16', 'Int32', 'UInt32',
                    'Int64', 'UInt64', 'Float32', 'Float64')

namespace = {'': "http://xml.opendap.org/ns/DAP/4.0#"}


def dap4_to_numpy_typemap(type_string):
    """
    This function takes a numpy dtype object
    and returns a dtype object that is compatible with
    the DAP2 specification.
    """
    dtype_str = pydap.lib.DAP4_TO_NUMPY_PARSER_TYPEMAP[type_string]
    return np.dtype(dtype_str)


def get_variables(node, prefix=''):
    variables = collections.OrderedDict()
    group_name = node.get('name')
    if group_name is None:
        return variables
    if node.tag != 'Dataset':
        prefix = prefix + '/' + group_name
    for subnode in node:
        if subnode.tag in dmr_atomic_types:
            name = subnode.get('name')
            if prefix != '':
                name = prefix + '/' + name
            variables[name] = {'element': subnode}
        variables.update(get_variables(subnode, prefix))
    return variables


def get_named_dimensions(node, prefix=''):
    dimensions = {}
    group_name = node.get('name')
    if group_name is None:
        return dimensions
    if node.tag != 'Dataset':
        prefix = prefix + '/' + group_name
    for subnode in node:
        if subnode.tag == 'Dimension':
            name = subnode.get('name')
            if prefix != '':
                name = prefix + '/' + name
            dimensions[name] = int(subnode.attrib['size'])
        dimensions.update(get_named_dimensions(subnode, prefix))
    return dimensions


def get_dtype(element):
    dtype = element.tag
    dtype = dap4_to_numpy_typemap(dtype)
    return dtype


def get_attributes(element):
    attributes = {}
    attribute_elements = element.findall('Attribute')
    for attribute_element in attribute_elements:
        name = attribute_element.get('name')
        value = attribute_element.find('Value').text
        attributes[name] = value
    return attributes


def get_dim_names(element):
    n_unnamed = 0
    # Not to be confused with dimensions
    dimension_elements = element.findall('Dim')
    dimensions = []
    for dimension_element in dimension_elements:
        name = dimension_element.get('name')
        if name is None:
            # We might have unnamed dimensions
            return dimensions
        if name.find('/', 1) == -1:
            # If this is a root Dimension, we remove the leading slash
            name = name.replace('/', '')
        dimensions.append(name)
    return dimensions


def get_dim_sizes(element):
    dimension_elements = element.findall('Dim')
    dimension_sizes = ()
    for dimension_element in dimension_elements:
        name = dimension_element.get('name')
        if name is None:
            size = int(dimension_element.get('size'))
            dimension_sizes += (size,)
    return dimension_sizes


def has_map(element):
    maps = element.findall('Map')
    if len(maps) > 0:
        return True
    else:
        return False


def dmr_to_dataset(dmr):
    """Return a dataset object from a DMR representation."""

    # Parse the DMR. First dropping the namespace
    dmr = re.sub(' xmlns="[^"]+"', '', dmr, count=1)
    dom_et = ET.fromstring(dmr)

    variables = get_variables(dom_et)
    named_dimensions = get_named_dimensions(dom_et)

    # Add size entry for dimension variables
    for name, size in named_dimensions.items():
        if name in variables:
            variables[name]['size'] = size

    # Bootstrap variables
    for name, variable in variables.items():
        variable['name'] = name
        variable['attributes'] = get_attributes(variable['element'])
        variable['dtype'] = get_dtype(variable['element'])
        variable['dims'] = get_dim_names(variable['element'])
        variable['has_map'] = has_map(variable['element'])
        variable['shape'] = get_dim_sizes(variable['element'])

    # Add size entry for dimension variables
    for name, size in named_dimensions.items():
        if name not in variables:
            # We might have dimensions that only have a declaration, so we need to add them to the variables
            variables[name] = {'name': name, 'size': size, 'dims': [name],
                               'dtype': 'int', 'has_map': False, 'attributes': {}, 'shape': ()}

    # Add shape element to variables
    for name, variable in variables.items():
        for dim in variable['dims']:
            variable['shape'] += (variables[dim]['size'],)

    # Convert the ordered dictionary to dataset
    dataset_name = dom_et.attrib['name']
    dataset = pydap.model.DatasetType(dataset_name)
    for name, variable in variables.items():
        data = DummyData(dtype=variable['dtype'], shape=variable['shape'])
        array = pydap.model.BaseType(name=variable['name'], data=data, dimensions=variable['dims'])
        if variable['has_map']:
            var = pydap.model.GridType(name=variable['name'])
            var[name] = array
            for dim in variable['dims']:
                var[dim] = copy.copy(dataset[dim])
        else:
            var = array
        var.attributes = variable['attributes']
        dataset[var.name] = var

    return dataset


class DMRParser:
    """A parser for the DMR."""

    def __init__(self, dmr):
        self.dmr = dmr


class DummyData(object):
    def __init__(self, dtype, shape):
        self.dtype = dtype
        self.shape = shape



"""pydap handler for NetCDF3/4 files."""

import os
import re
import time
from stat import ST_MTIME
from email.utils import formatdate
import numpy as np

from pkg_resources import get_distribution

from ...model import DatasetType, GridType, BaseType
from ..lib import BaseHandler
from ...exceptions import OpenFileError
from ...pycompat import suppress

from collections import OrderedDict

# Check for netCDF4 presence:
with suppress(ImportError):
    try:
        from netCDF4 import Dataset as netcdf_file

        def attrs(var):
            return dict((k, getattr(var, k)) for k in var.ncattrs())
    except ImportError:
        from scipy.io.netcdf import netcdf_file

        def attrs(var):
            return var._attributes


class NetCDFHandler(BaseHandler):

    """A simple handler for NetCDF files.
    Here's a standard dataset for testing sequential data:
    """

    __version__ = get_distribution("pydap").version
    extensions = re.compile(r"^.*\.(nc|cdf)$", re.IGNORECASE)

    def __init__(self, filepath):
        BaseHandler.__init__(self)

        self.filepath = filepath
        try:
            with netcdf_file(self.filepath, 'r') as source:
                self.additional_headers.append(('Last-modified',
                                               (formatdate(
                                                time.mktime(
                                                    time.localtime(
                                                        os.stat(filepath)
                                                        [ST_MTIME])
                                                        )))))

                # shortcuts
                vars = source.variables
                dims = source.dimensions

                # build dataset
                name = os.path.split(filepath)[1]
                self.dataset = DatasetType(name,
                                           attributes=dict(
                                                      NC_GLOBAL=attrs(source)))
                for dim in dims:
                    if dims[dim] is None:
                        self.dataset.attributes['DODS_EXTRA'] = {
                            'Unlimited_Dimension': dim,
                        }
                        break

                # add grids
                grids = [var for var in vars if var not in dims]
                for grid in grids:
                    self.dataset[grid] = GridType(grid, attrs(vars[grid]))
                    # add array
                    self.dataset[grid][grid] = BaseType(grid,
                                                        LazyVariable(
                                                            source,
                                                            grid,
                                                            grid,
                                                            self.filepath),
                                                        vars[grid].dimensions,
                                                        attrs(vars[grid]))
                    # add maps
                    for dim in vars[grid].dimensions:
                        try:
                            data = vars[dim][:]
                            attributes = attrs(vars[dim])
                        except KeyError:
                            data = np.arange(dims[dim].size, dtype='i')
                            attributes = None
                        self.dataset[grid][dim] = BaseType(dim, data,
                                                           None,
                                                           attributes)

                # add dims
                for dim in dims:
                    try:
                        data = vars[dim][:]
                        attributes = attrs(vars[dim])
                    except KeyError:
                        data = np.arange(dims[dim].size, dtype='i')
                        attributes = None
                    self.dataset[dim] = BaseType(dim, data,
                                                 None,
                                                 attributes)
        except Exception as exc:
            raise
            message = 'Unable to open file %s: %s' % (filepath, exc)
            raise OpenFileError(message)


class LazyVariable:
    def __init__(self, source, name, path, filepath):
        self.filepath = filepath
        self.path = path
        var = source[self.path]
        self.dimensions = var._getdims()
        self.dtype = np.dtype(var.dtype)
        self.datatype = var.dtype
        self.ndim = len(var.dimensions)
        self._shape = var.shape
        self._reshape = var.shape
        self.scale = True
        self.name = name
        self.size = np.prod(self.shape)
        self._attributes = dict((attr, var.getncattr(attr))
                                for attr in var.ncattrs())
        return

    def chunking(self):
        return 'contiguous'

    def filters(self):
        return None

    def get_var_chunk_cache(self):
        raise NotImplementedError('get_var_chunk_cache is not implemented')
        return

    def ncattrs(self):
        return self._attributes

    def getncattr(self, attr):
        return self._attributes[attr]

    def __getattr__(self, name):
        # from netcdf4-python
        # if name in _private_atts, it is stored at the python
        # level and not in the netCDF file.
        if name.startswith('__') and name.endswith('__'):
            # if __dict__ requested, return a dict with netCDF attributes.
            if name == '__dict__':
                names = self.ncattrs()
                values = []
                for name in names:
                    values.append(self._attributes[name])
                return OrderedDict(zip(names, values))
            else:
                raise AttributeError
        else:
            return self.getncattr(name)

    def getValue(self):
        return self[...]

    def __array__(self):
        return self[...]

    def __getitem__(self, key):
        with netcdf_file(self.filepath, 'r') as source:
            # Avoid applying scale_factor, see
            # https://github.com/pydap/pydap/issues/190
            source.set_auto_scale(False)
            return (np.asarray(source[self.path][key])
                    .astype(self.dtype).reshape(self._reshape))

    def reshape(self, *args):
        if len(args) > 1:
            self._reshape = args
        else:
            self._reshape = args
        return self

    @property
    def shape(self):
        return self._shape

    def __len__(self):
        if not self.shape:
            raise TypeError('len() of unsized object')
        else:
            return self.shape[0]

    def _getdims(self):
        return self.dimensions


if __name__ == "__main__":
    import sys
    from werkzeug.serving import run_simple

    application = NetCDFHandler(sys.argv[1])
    run_simple('localhost', 8001, application, use_reloader=True)



"""
Test the follow redirects and handling of more complex routing situations
"""

import requests
from webob.request import Request
import requests_mock
from pydap.net import create_request


def test_redirect():
    """ Test that redirection is handled properly """
    mock_url = 'http://www.test.com/'
    # mock_url is redirected to https:
    redirect_url = 'http://www.test2.com/'
    with requests_mock.Mocker() as m:
        m.register_uri('HEAD', mock_url, text='resp2', status_code=200)
        req = create_request(mock_url)
        assert len(m.request_history) == 1
        assert isinstance(req, Request)
        assert req.headers['Host'] == 'www.test.com:80'

    # Without session:
    with requests_mock.Mocker() as m:
        m.register_uri('HEAD', mock_url, text='resp1', status_code=301,
                       headers={'Location': redirect_url})
        m.register_uri('HEAD', redirect_url, text='resp2', status_code=200)
        req = create_request(mock_url)
        # Ensure follow redirect:
        assert len(m.request_history) == 2
        assert isinstance(req, Request)
        assert req.headers['Host'] == 'www.test2.com:80'

    # With session:
    with requests_mock.Mocker() as m:
        m.register_uri('HEAD', mock_url, text='resp1', status_code=301,
                       headers={'Location': redirect_url})
        m.register_uri('HEAD', redirect_url, text='resp2', status_code=200)
        req = create_request(mock_url, session=requests.Session())
        # Ensure follow redirect:
        assert len(m.request_history) == 2
        assert isinstance(req, Request)
        assert req.headers['Host'] == 'www.test2.com:80'



#
# Test loading a URL from the EarthData Cloud (EDC) OPeNDAP server using the
# new (8/18/22) PyDAP DAP4 support. This dataset will not load using DAP2.
#

import pydap.client
import requests
import configparser
import time
import logging

logging.basicConfig(filename='dap4_access_test.log',
                    filemode='w',
                    level=logging.DEBUG,
                    format='%(asctime)s %(levelname)s %(name)s : %(message)s')

logger = logging.getLogger(__name__)


class SessionEarthData(requests.Session):
    AUTH_HOST = 'urs.earthdata.nasa.gov'

    def __init__(self, username, password):
        super().__init__()
        self.auth = (username, password)

    def rebuild_auth(self, prepared_request, response):
        headers = prepared_request.headers
        url = prepared_request.url
        if 'Authorization' in headers:
            original_parsed = requests.utils.urlparse(response.request.url)
            redirect_parsed = requests.utils.urlparse(url)
            if (original_parsed.hostname != redirect_parsed.hostname) and \
                    redirect_parsed.hostname != self.AUTH_HOST and \
                    original_parsed.hostname != self.AUTH_HOST:
                del headers['Authorization']
        return


def main():
    config = configparser.ConfigParser()
    config.read('user.config')
    username = config['user']['user']
    password = config['user']['pwd']
    session = SessionEarthData(username=username, password=password)

    dap4_schema = 'dap4'

    host = 'opendap.earthdata.nasa.gov'
    path = '/collections/C1996881146-POCLOUD/granules/'
    dataset = '20220531090000-JPL-L4_GHRSST-SSTfnd-MUR-GLOB-v02.0-fv04.1'

    dap4_url = f'{dap4_schema}://{host}{path}{dataset}'

    # open the dap4 URL
    print(f'Open this URL: {dap4_url}')
    i = 0
    while True:
        start = time.time()
        pydap_ds = pydap.client.open_url(dap4_url, session=session)

        print(f'The attributes:')
        print(pydap_ds['sea_ice_fraction'].attributes)
        i += 1
        print(f'Call {i}')
        variable = pydap_ds['sea_ice_fraction'][0, 1700:1799:10, 1800:1900:10]
        print(f'Time for request: {time.time()-start}')
        print(f'A subset of the "sea_ice_fraction" variable')
        print(variable.data)


if __name__ == '__main__':
    main()




"""A file-based pydap server running on Gunicorn.

Usage:
  pydap [options]

Options:
  -h --help                     Show this help message and exit
  --version                     Show version
  -i --init DIR                 Create directory with templates
  -b ADDRESS --bind ADDRESS     The ip to listen to [default: 127.0.0.1]
  -p PORT --port PORT           The port to connect [default: 8001]
  -d DIR --data DIR             The directory with files [default: .]
  -t DIR --templates DIR        The directory with templates
  --worker-class=CLASS          Gunicorn worker class [default: sync]

"""

import os
import re
import mimetypes
from datetime import datetime
import shutil

from jinja2 import Environment, PackageLoader, FileSystemLoader, ChoiceLoader
from webob import Response
from webob.dec import wsgify
from webob.exc import HTTPNotFound, HTTPForbidden
from webob.static import FileApp, DirectoryApp
import pkg_resources
from six.moves.urllib.parse import unquote
from six import string_types

from ..lib import __version__
from ..handlers.lib import get_handler, load_handlers
from ..exceptions import ExtensionNotSupportedError
from .ssf import ServerSideFunctions


class DapServer(object):

    """A directory app that creates file listings and handle DAP requests."""

    def __init__(self, path, templates=None):
        self.path = os.path.abspath(path)

        # the default loader reads templates from the package
        loaders = [PackageLoader("pydap.wsgi", "templates")]

        # optionally, the user can also specify a template directory that will
        # override the default templates; this should have precedence over the
        # default templates
        if templates is not None:
            loaders.insert(0, FileSystemLoader(templates))

        # set the rendering environment; this is also used by pydap responses
        # that need to render templates (like HTML, WMS, KML, etc.)
        self.env = Environment(loader=ChoiceLoader(loaders))
        self.env.filters["datetimeformat"] = datetimeformat
        self.env.filters["datetimeformat_iso"] = datetimeformat_iso
        self.env.filters["unquote"] = unquote

        # cache available handlers, so we don't need to load them every request
        self.handlers = load_handlers()

    @wsgify
    def __call__(self, req):
        """WSGI application callable.

        Returns either a file download, directory listing or DAP response.

        """
        path = os.path.abspath(
            os.path.join(self.path, *req.path_info.split("/")))

        if not path.startswith(self.path):
            return HTTPForbidden()
        if path.endswith('catalog.xml'):
            return self.index(os.path.dirname(path), req, catalog=True)
        elif os.path.exists(path):
            if os.path.isdir(path):
                return self.index(path, req)
            else:
                return FileApp(path)

        # strip DAP extension (``.das``, eg) and see if the file exists
        base, ext = os.path.splitext(path)
        if os.path.isfile(base):
            req.environ["pydap.jinja2.environment"] = self.env
            app = ServerSideFunctions(get_handler(base, self.handlers))
            return req.get_response(app)
        else:
            return HTTPNotFound(comment=path)

    def index(self, directory, req, catalog=False):
        """Return a directory listing."""

        template_name = 'index.html'
        response_content_type = 'text/html'
        content = [
            os.path.join(directory, name) for name in os.listdir(directory)]

        files = [{
            "name": os.path.split(path)[1],
            "size": os.path.getsize(path),
            "last_modified": datetime.fromtimestamp(os.path.getmtime(path)),
            "supported": supported(path, self.handlers),
        } for path in content if os.path.isfile(path)]
        files.sort(key=lambda d: alphanum_key(d["name"]))

        directories = [{
            "name": os.path.split(path)[1],
            "last_modified": datetime.fromtimestamp(os.path.getmtime(path)),
        } for path in content if os.path.isdir(path)]
        directories.sort(key=lambda d: alphanum_key(d["name"]))

        tokens = req.path_info.split("/")[1:]
        breadcrumbs = [{
            "url": "/".join([req.application_url] + tokens[:i+1]),
            "title": token,
        } for i, token in enumerate(tokens) if token]

        context = {
            "root": req.application_url,
            "location": req.path_url,
            "breadcrumbs": breadcrumbs,
            "directories": directories,
            "files": files,
            "version": __version__,
        }

        if catalog:
            context['location'] = req.path.replace('catalog.xml', '')
            template_name = 'catalog.xml'
            response_content_type = 'application/xml'

        template = self.env.get_template(template_name)
        return Response(
            body=template.render(context),
            content_type=response_content_type,
            charset="utf-8")


def supported(filepath, handlers=None):
    """Test if a file has a corresponding handler.

    Returns a boolean.

    """
    try:
        get_handler(filepath, handlers, instantiate=False)
        return True
    except ExtensionNotSupportedError:
        return False


def alphanum_key(s):
    """Parse a string, returning a list of string and number chunks.

        >>> alphanum_key("z23a")
        ['z', 23, 'a']

    Useful for sorting names in a natural way.

    From http://nedbatchelder.com/blog/200712.html#e20071211T054956

    """
    def tryint(s):
        try:
            return int(s)
        except Exception:
            return s
    return [tryint(c) for c in re.split('([0-9]+)', s)]


def datetimeformat(value, format='%Y-%m-%d %H:%M:%S'):
    """Return a formatted datetime object."""
    return value.strftime(format)


def datetimeformat_iso(value):
    """Return a formatted datetime object as ISO8601 representation."""
    return value.isoformat()


class StaticMiddleware(object):

    """WSGI middleware for static assets.

    The assets can be either specified as a directory, or retrieved from a
    Python package. Inspired by ``werkezeug.wsgi.SharedDataMiddleware``.

    """

    def __init__(self, app, static):
        self.app = app
        self.static = static

    @wsgify
    def __call__(self, req):
        if req.path_info_peek() != "static":
            return req.get_response(self.app)

        # strip "/static"
        req.path_info_pop()

        # statically serve the directory
        if isinstance(self.static, string_types):
            return req.get_response(DirectoryApp(self.static))

        # otherwise, load resource from package
        package, resource_path = self.static
        resource = os.path.join(resource_path, *req.path_info.split('/'))
        if not pkg_resources.resource_exists(package, resource):
            return HTTPNotFound(req.path_info)

        content_type, content_encoding = mimetypes.guess_type(resource)
        return Response(
            body=pkg_resources.resource_string(package, resource),
            content_type=content_type,
            content_encoding=content_encoding)


def init(directory):
    """Create directory with default templates."""
    # copy main templates
    templates = pkg_resources.resource_filename("pydap.wsgi", "templates")
    shutil.copytree(templates, directory)

    # copy templates from HTML response
    for resource in pkg_resources.resource_listdir(
            "pydap.responses.html", "templates"):
        path = pkg_resources.resource_filename(
            "pydap.responses.html", "templates/{0}".format(resource))
        shutil.copy(path, directory)


def main():  # pragma: no cover
    """Run server from the command line."""
    import multiprocessing

    from docopt import docopt
    from gunicorn.app.pasterapp import PasterServerApplication

    arguments = docopt(__doc__, version="pydap %s" % __version__)

    # init templates?
    if arguments["--init"]:
        init(arguments["--init"])
        return

    # create pydap app
    data, templates = arguments["--data"], arguments["--templates"]
    app = DapServer(data, templates)

    # configure app so that is reads static assets from the template directory
    # or from the package
    if templates and os.path.exists(os.path.join(templates, "static")):
        static = os.path.join(templates, "static")
    else:
        static = ("pydap.wsgi", "templates/static")
    app = StaticMiddleware(app, static)

    # configure WSGI server
    workers = multiprocessing.cpu_count() * 2 + 1
    PasterServerApplication(
        app,
        host=arguments["--bind"],
        port=int(arguments["--port"]),
        workers=workers,
        worker_class=arguments["--worker-class"]).run()


if __name__ == "__main__":
    main()



"""A DDS parser."""

import re

import numpy as np

from . import SimpleParser
from ..model import (DatasetType, BaseType,
                     SequenceType, StructureType,
                     GridType)
from ..lib import (quote, LOWER_DAP2_TO_NUMPY_PARSER_TYPEMAP)

constructors = ('grid', 'sequence', 'structure')
name_regexp = r'[\w%!~"\'\*-]+'


def DAP2_parser_typemap(type_string):
    """
    This function takes a numpy dtype object
    and returns a dtype object that is compatible with
    the DAP2 specification.
    """
    dtype_str = LOWER_DAP2_TO_NUMPY_PARSER_TYPEMAP[type_string.lower()]
    return np.dtype(dtype_str)


class DDSParser(SimpleParser):
    """A parser for the DDS."""

    def __init__(self, dds):
        super(DDSParser, self).__init__(dds, re.IGNORECASE)
        self.dds = dds

    def consume(self, regexp):
        """Consume and return a token."""
        token = super(DDSParser, self).consume(regexp)
        self.buffer = self.buffer.lstrip()
        return token

    def parse(self):
        """Parse the DAS, returning a dataset."""
        dataset = DatasetType('nameless')

        self.consume('dataset')
        self.consume('{')
        while not self.peek('}'):
            var = self.declaration()
            dataset[var.name] = var
        self.consume('}')

        dataset.name = quote(self.consume('[^;]+'))
        dataset._set_id(dataset.name)
        self.consume(';')

        return dataset

    def declaration(self):
        """Parse and return a declaration."""
        token = self.peek(r'\w+').lower()

        map = {
            'grid':      self.grid,
            'sequence':  self.sequence,
            'structure': self.structure,
        }
        method = map.get(token, self.base)
        return method()

    def base(self):
        """Parse a base variable, returning a ``BaseType``."""
        data_type_string = self.consume(r'\w+')

        parser_dtype = DAP2_parser_typemap(data_type_string)
        name = quote(self.consume(r'[^;\[]+'))

        shape, dimensions = self.dimensions()
        self.consume(r';')

        data = DummyData(parser_dtype, shape)
        var = BaseType(name, data, dimensions=dimensions)

        return var

    def dimensions(self):
        """Parse variable dimensions, returning tuples of dimensions/names."""
        shape = []
        names = []
        while not self.peek(';'):
            self.consume(r'\[')
            token = self.consume(name_regexp)
            if self.peek('='):
                names.append(token)
                self.consume('=')
                token = self.consume(r'\d+')
            shape.append(int(token))
            self.consume(r'\]')
        return tuple(shape), tuple(names)

    def sequence(self):
        """Parse a DAS sequence, returning a ``SequenceType``."""
        sequence = SequenceType('nameless')
        self.consume('sequence')
        self.consume('{')

        while not self.peek('}'):
            var = self.declaration()
            sequence[var.name] = var
        self.consume('}')

        sequence.name = quote(self.consume('[^;]+'))
        self.consume(';')
        return sequence

    def structure(self):
        """Parse a DAP structure, returning a ``StructureType``."""
        structure = StructureType('nameless')
        self.consume('structure')
        self.consume('{')

        while not self.peek('}'):
            var = self.declaration()
            structure[var.name] = var
        self.consume('}')

        structure.name = quote(self.consume('[^;]+'))
        self.consume(';')

        return structure

    def grid(self):
        """Parse a DAP grid, returning a ``GridType``."""
        grid = GridType('nameless')
        self.consume('grid')
        self.consume('{')

        self.consume('array')
        self.consume(':')
        array = self.base()
        grid[array.name] = array

        self.consume('maps')
        self.consume(':')
        while not self.peek('}'):
            var = self.base()
            grid[var.name] = var
        self.consume('}')

        grid.name = quote(self.consume('[^;]+'))
        self.consume(';')

        return grid


def dds_to_dataset(dds):
    """Return a dataset object from a DDS representation."""
    return DDSParser(dds).parse()


class DummyData(object):
    def __init__(self, dtype, shape):
        self.dtype = dtype
        self.shape = shape



"""Test IterData, which makes iterables behave like Numpy structured arrays."""

import numpy as np
from pydap.handlers.lib import IterData
from pydap.model import DatasetType, SequenceType, BaseType
import pytest


@pytest.fixture
def simple_array():
    arr = np.array([
            (0, 1, 10.),
            (1, 2, 20.),
            (2, 3, 30.),
            (3, 4, 40.),
            (4, 5, 50.),
            (5, 6, 60.),
            (6, 7, 70.),
            (7, 8, 80.),
        ], dtype=[('byte', 'b'), ('int', 'i4'), ('float', 'f4')])
    return arr


@pytest.fixture
def simple_object(simple_array):
    # add sequence and children for each column
    name = 'nameless'
    dataset = DatasetType(name)
    seq = dataset['sequence'] = SequenceType('sequence')
    for var in simple_array.dtype.names:
        seq[var] = BaseType(var)

    obj = IterData([(0, 1, 10.),
                    (1, 2, 20.),
                    (2, 3, 30.),
                    (3, 4, 40.),
                    (4, 5, 50.),
                    (5, 6, 60.),
                    (6, 7, 70.),
                    (7, 8, 80.)],
                   seq)
    return obj


def test_iter(simple_array, simple_object):
    assert ([tuple(row)
             for row in simple_array] ==
            [(0, 1, 10.), (1, 2, 20.),
             (2, 3, 30.), (3, 4, 40.),
             (4, 5, 50.), (5, 6, 60.),
             (6, 7, 70.), (7, 8, 80.)])
    assert (list(simple_array["byte"]) ==
            [0, 1, 2, 3, 4, 5, 6, 7])

    assert ([tuple(row) for row in simple_object] ==
            [(0, 1, 10.), (1, 2, 20.),
             (2, 3, 30.), (3, 4, 40.),
             (4, 5, 50.), (5, 6, 60.),
             (6, 7, 70.), (7, 8, 80.)])
    assert (list(simple_object["byte"]) ==
            [0, 1, 2, 3, 4, 5, 6, 7])


def test_dtype(simple_array, simple_object):
    assert (
        simple_array.dtype ==
        np.dtype([('byte', 'i1'), ('int', '<i4'), ('float', '<f4')]))
    assert (simple_array["int"].dtype == "<i4")

    assert (
        simple_object.dtype ==
        np.dtype([('byte', '<i8'), ('int', '<i8'), ('float', '<f8')]))
    assert (simple_object["int"].dtype == "<i8")


def test_selection(simple_array, simple_object):
    selection = simple_array[simple_array["byte"] > 3]
    assert ([tuple(row) for row in selection] ==
            [(4, 5, 50.), (5, 6, 60.),
             (6, 7, 70.), (7, 8, 80.)])

    selection = simple_object[simple_object["byte"] > 3]
    assert ([tuple(row) for row in selection] ==
            [(4, 5, 50.), (5, 6, 60.),
             (6, 7, 70.), (7, 8, 80.)])


def test_projection(simple_array, simple_object):
    projection = simple_array[1::2]
    assert ([tuple(row) for row in projection] ==
            [(1, 2, 20.), (3, 4, 40.),
             (5, 6, 60.), (7, 8, 80.)])
    assert (
        list(simple_array["byte"][1::2]) == [1, 3, 5, 7])
    assert (
        list(simple_array[1::2]["byte"]) == [1, 3, 5, 7])

    projection = simple_object[1::2]
    assert ([tuple(row) for row in projection] ==
            [(1, 2, 20.), (3, 4, 40.),
             (5, 6, 60.), (7, 8, 80.)])
    assert (
        list(simple_object["byte"][1::2]) == [1, 3, 5, 7])
    assert (
        list(simple_object[1::2]["byte"]) == [1, 3, 5, 7])


def test_combined(simple_array, simple_object):
    filtered = simple_array[simple_array["byte"] > 1]
    filtered = filtered[filtered["byte"] < 6]
    filtered = filtered[::2]
    assert ([tuple(row) for row in filtered] ==
            [(2, 3, 30.), (4, 5, 50.)])

    filtered = simple_object[simple_object["byte"] > 1]
    filtered = filtered[filtered["byte"] < 6]
    filtered = filtered[::2]
    assert ([tuple(row) for row in filtered] ==
            [(2, 3, 30.), (4, 5, 50.)])


def test_combined_other(simple_array, simple_object):
    filtered = simple_array[['int', 'float']][simple_array["byte"] < 2]
    assert ([tuple(row) for row in filtered] ==
            [(1, 10.), (2, 20.)])

    filtered = simple_object[['int', 'float']][simple_object["byte"] < 2]
    assert ([tuple(row) for row in filtered] ==
            [(1, 10.), (2, 20.)])


@pytest.fixture
def nested_data():
    shallow_data = [(1, 1, 1), (2, 4, 4),
                    (3, 6, 9), (4, 8, 16)]
    deep_data = [[(10, 11, 12), (21, 22, 23)],
                 [(15, 16, 17)],
                 [],
                 [(31, 32, 33), (41, 42, 43),
                  (51, 52, 53), (61, 62, 63)]]
    nested = [x + (deep_data[x_id],)
              for x_id, x
              in enumerate(shallow_data)]
    return nested


@pytest.fixture
def nested_dtype():
    dtype = np.dtype([('a', '<i8'), ('b', '<i8'),
                      ('c', '<i8'),
                      ('d', np.dtype([('e', '<i8'),
                                      ('f', '<i8'),
                                      ('g', '<i8')]))])
    return dtype


@pytest.fixture
def nested_object(nested_data):
    name = 'nameless'
    dataset = DatasetType(name)
    seq = dataset['nested'] = SequenceType('nested')
    for var in ['a', 'b', 'c']:
        seq[var] = BaseType(var)
    seq['d'] = SequenceType('d')
    for var in ['e', 'f', 'g']:
        seq['d'][var] = BaseType(var)

    nested = IterData(nested_data, seq)
    return nested


def test_nested_iter(nested_data, nested_object):
    assert ([tuple(row) for row in nested_object] ==
            nested_data)


def test_nested_iter_child(nested_data, nested_object):
    expected = [row[0] for row in nested_data]
    assert list(nested_object["a"]) == expected


def test_nested_iter_nested_sequence(nested_data, nested_object):
    expected = [row[3] for row in nested_data]
    assert list(nested_object["d"]) == expected


def test_nested_iter_nested_deep_child(nested_data, nested_object):
    expected = [[col[0] for col in row[3]]
                for row in nested_data]
    assert list(nested_object['d']['e']) == expected


def test_nested_dtype(nested_dtype, nested_object):
    assert nested_object.dtype == nested_dtype


def test_nested_selection(nested_data, nested_object):
    selection = nested_object[nested_object["a"] > 2]
    expected = [tuple(row) for row in nested_data
                if row[0] > 2]
    assert [tuple(row) for row in selection] == expected


def test_nested_projection(nested_data, nested_object):
    projection = nested_data[1::2]
    expected = [tuple(row) for row_id, row
                in enumerate(nested_data)
                if row_id in range(1, len(nested_data), 2)]
    assert [tuple(row) for row in projection] == expected

    projection = nested_object[1::2]
    expected = [tuple(row) for row_id, row
                in enumerate(nested_data)
                if row_id in range(1, len(nested_data), 2)]
    assert [tuple(row) for row in projection] == expected


def test_nested_combined(nested_data, nested_object):
    filtered = [tuple(row) for row in nested_data
                if row[0] > 2]
    filtered = [tuple(row) for row in filtered
                if row[0] < 4]
    filtered = filtered[::2]
    assert ([tuple(row) for row in filtered] ==
            [(3, 6, 9, [])])

    filtered = nested_object[nested_object["a"] > 2]
    filtered = filtered[filtered["a"] < 4]
    filtered = filtered[::2]
    assert ([tuple(row) for row in filtered] ==
            [(3, 6, 9, [])])



"""Datasets for testing.

This module defines a few datasets for testing, covering the whole DAP data
model.

"""
import os

import numpy as np

from pydap.model import (DatasetType, BaseType,
                         SequenceType, GridType, StructureType)
from pydap.handlers.lib import IterData
from pydap.client import open_file

from collections import OrderedDict

# Note that DAP2 does not support signed bytes (signed 8bits integers).

# A very simple sequence: flat and with no strings. This sequence can be mapped
# directly to a Numpy structured array, and can be easily encoded and decoded
# in the DAP spec.
VerySimpleSequence = DatasetType("VerySimpleSequence")
VerySimpleSequence["sequence"] = SequenceType("sequence")
VerySimpleSequence["sequence"]["byte"] = BaseType("byte")
VerySimpleSequence["sequence"]["int"] = BaseType("int")
VerySimpleSequence["sequence"]["float"] = BaseType("float")
VerySimpleSequence["sequence"].data = np.array([
    (0, 1, 10.),
    (1, 2, 20.),
    (2, 3, 30.),
    (3, 4, 40.),
    (4, 5, 50.),
    (5, 6, 60.),
    (6, 7, 70.),
    (7, 8, 80.),
    ], dtype=[('byte', np.ubyte), ('int', 'i4'), ('float', 'f4')])


# A nested sequence.
NestedSequence = DatasetType("NestedSequence")
NestedSequence["location"] = SequenceType("location")
NestedSequence["location"]["lat"] = BaseType("lat")
NestedSequence["location"]["lon"] = BaseType("lon")
NestedSequence["location"]["elev"] = BaseType("elev")
NestedSequence["location"]["time_series"] = SequenceType("time_series")
NestedSequence["location"]["time_series"]["time"] = BaseType("time")
NestedSequence["location"]["time_series"]["slp"] = BaseType("slp")
NestedSequence["location"]["time_series"]["wind"] = BaseType("wind")
NestedSequence["location"].data = IterData([
    (1, 1, 1, [(10, 11, 12), (21, 22, 23)]),
    (2, 4, 4, [(15, 16, 17)]),
    (3, 6, 9, []),
    (4, 8, 16, [(31, 32, 33), (41, 42, 43), (51, 52, 53), (61, 62, 63)]),
    ], NestedSequence.location)


# A simple array with bytes, strings and shorts. These types require special
# encoding for the DODS response.
SimpleArray = DatasetType("SimpleArray")
SimpleArray["byte"] = BaseType("byte", np.arange(5, dtype=np.ubyte))
SimpleArray["string"] = BaseType("string", np.array(["one", "two"]))
SimpleArray["short"] = BaseType("short", np.array(1, dtype="h"))


DODS = os.path.join(os.path.dirname(__file__), 'data/rainfall_time_malaysia.cdp.dods')
DAS = os.path.join(os.path.dirname(__file__), 'data/rainfall_time_malaysia.cdp.das')
dapper = open_file(DODS, DAS)


# dataset from http://test.opendap.org:8080/dods/dts/D1.asc
D1 = DatasetType('EOSDB.DBO', type='Drifters')
D1['Drifters'] = SequenceType('Drifters')
D1['Drifters']['instrument_id'] = BaseType('instrument_id')
D1['Drifters']['location'] = BaseType('location')
D1['Drifters']['latitude'] = BaseType('latitude')
D1['Drifters']['longitude'] = BaseType('longitude')
D1.Drifters.data = np.array(np.rec.fromrecords(list(zip(
    [u"This is a data test string (pass {0}).".format(1+i*2)
     for i in range(5)],
    [u"This is a data test string (pass {0}).".format(i*2) for i in range(5)],
    [1000.0, 999.95, 999.80, 999.55, 999.20],
    [999.95, 999.55, 998.75, 997.55, 995.95])),
    names=list(D1.Drifters.keys())))


# testing structures
SimpleStructure = DatasetType('SimpleStructure')
SimpleStructure['types'] = StructureType(
    name='types',
    key="value",
    nested=OrderedDict([
        ("string", "bar"),
        ("list", [42, 43]),
        ("array", np.array(1)),
        ("float", 1000.0),
    ]))
SimpleStructure['types']['b'] = BaseType('b', np.array(-10, np.byte))
SimpleStructure['types']['ub'] = BaseType('ub', np.array(10, np.ubyte))
SimpleStructure['types']['i32'] = BaseType('i32', np.array(-10, np.int32))
SimpleStructure['types']['ui32'] = BaseType('ui32', np.array(10, np.uint32))
SimpleStructure['types']['i16'] = BaseType('i16', np.array(-10, np.int16))
SimpleStructure['types']['ui16'] = BaseType('ui16', np.array(10, np.uint16))
SimpleStructure['types']['f32'] = BaseType('f32', np.array(100.0, np.float32))
SimpleStructure['types']['f64'] = BaseType('f64', np.array(1000., np.float64))
SimpleStructure['types']['s'] = BaseType(
    's', np.array("This is a data test string (pass 0)."))
SimpleStructure['types']['u'] = BaseType('u', np.array("http://www.dods.org"))
SimpleStructure['types']['U'] = BaseType('U', np.array(u"test unicode",
                                                       str))


# test grid
rain = DatasetType('test')
rain['rain'] = GridType('rain')
rain['rain']['rain'] = BaseType(
    'rain', np.arange(6).reshape(2, 3), dimensions=('y', 'x'))
rain['rain']['x'] = BaseType('x', np.arange(3), units='degrees_east')
rain['rain']['y'] = BaseType('y', np.arange(2), units='degrees_north')


# test for ``bounds`` function
bounds = DatasetType('test')
bounds['sequence'] = SequenceType('sequence')
bounds['sequence']['lon'] = BaseType('lon', axis='X')
bounds['sequence']['lat'] = BaseType('lat', axis='Y')
bounds['sequence']['depth'] = BaseType('depth', axis='Z')
bounds['sequence']['time'] = BaseType(
    'time', axis='T', units="days since 1970-01-01")
bounds['sequence']['measurement'] = BaseType('measurement')
bounds.sequence.data = np.array(np.rec.fromrecords([
    (100, -10, 0, -1, 42),
    (200, 10, 500, 1, 43),
], names=list(bounds.sequence.keys())))


# test for density
ctd = DatasetType('ctd')
ctd['cast'] = SequenceType('cast')
ctd['cast']['temperature'] = BaseType('temperature')
ctd['cast']['salinity'] = BaseType('salinity')
ctd['cast']['pressure'] = BaseType('pressure')
ctd.cast.data = np.array(np.rec.fromrecords([
    (21, 35, 0),
    (15, 35, 100),
], names=list(ctd.cast.keys())))

# a simple sequence, simulating a CTD profile
SimpleSequence = DatasetType(
    "SimpleSequence", description="A simple sequence for testing.",
    nested={"value": 42})
SimpleSequence["cast"] = SequenceType("cast")
SimpleSequence["cast"]["id"] = BaseType("id")
SimpleSequence["cast"]["lon"] = BaseType("lon", axis="X")
SimpleSequence["cast"]["lat"] = BaseType("lat", axis="Y")
SimpleSequence["cast"]["depth"] = BaseType("depth", axis="Z")
SimpleSequence["cast"]["time"] = BaseType(
    "time", axis="T", units="days since 1970-01-01")
SimpleSequence["cast"]["temperature"] = BaseType("temperature")
SimpleSequence["cast"]["salinity"] = BaseType("salinity")
SimpleSequence["cast"]["pressure"] = BaseType("pressure")
SimpleSequence["cast"].data = np.array(np.rec.fromrecords([
    ("1", 100, -10,   0, -1, 21, 35,   0),
    ("2", 200,  10, 500,  1, 15, 35, 100),
], names=list(SimpleSequence['cast'].keys())))
# SimpleSequence["cast"].data = IterData([
#     ("1", 100, -10,   0, -1, 21, 35,   0),
#     ("2", 200,  10, 500,  1, 15, 35, 100),
# ], SimpleSequence.cast)

# a simple grid
SimpleGrid = DatasetType(
    "SimpleGrid", description="A simple grid for testing.")
SimpleGrid["SimpleGrid"] = GridType("SimpleGrid")
SimpleGrid["SimpleGrid"]["SimpleGrid"] = BaseType(
    "SimpleGrid", np.arange(6).reshape(2, 3), dimensions=("y", "x"))
SimpleGrid["x"] = SimpleGrid["SimpleGrid"]["x"] = BaseType(
    "x", np.arange(3), axis="X", units="degrees_east")
SimpleGrid["y"] = SimpleGrid["SimpleGrid"]["y"] = BaseType(
    "y", np.arange(2), axis="Y", units="degrees_north")


# a faulty grid
FaultyGrid = DatasetType(
    "FaultyGrid", description="A faulty grid for testing.")
FaultyGrid["FaultyGrid"] = GridType("FaultyGrid")
FaultyGrid["FaultyGrid"]["FaultyGrid"] = BaseType(
    "FaultyGrid", np.arange(6).reshape(2, 3), dimensions=("y", "x"))
FaultyGrid["x"] = FaultyGrid["FaultyGrid"]["x"] = BaseType(
    "x", np.arange(3), axis="X", code=1)
FaultyGrid["y"] = FaultyGrid["FaultyGrid"]["y"] = BaseType(
    "y", np.arange(2), axis="Y", code=np.int32([]))



from pydap.client import open_url
from pydap.cas import urs
import pydap.net
import requests
import os
import pytest


url = ('https://goldsmr3.gesdisc.eosdis.nasa.gov/opendap/'
       'MERRA_MONTHLY/MAIMCPASM.5.2.0/1979/'
       'MERRA100.prod.assim.instM_3d_asm_Cp.197901.hdf')
test_url = url + '.dods?SLP[0:1:0][0:1:10][0:1:10]'
test_url_2 = url + '.dods?PS[0:1:0][0:1:10][0:1:10]'


@pytest.mark.auth
@pytest.mark.prod_url
@pytest.mark.skipif(not (os.environ.get('USERNAME_URS') and
                         os.environ.get('PASSWORD_URS')),
                    reason=('Without auth credentials, '
                            'this test cannot work'))
def test_basic_urs_auth():
    """
    Set up PyDAP to use the URS request() function.

    The intent here is to ensure that pydap.net is able to
    open and url if and only if requests is able to
    open the same url.
    """
    session = urs.setup_session(os.environ.get('USERNAME_URS'),
                                os.environ.get('PASSWORD_URS'),
                                check_url=url)

    # Check that the requests library can access the link:
    res = requests.get(test_url, cookies=session.cookies)
    assert(res.status_code == 200)
    res.close()

    # Check that the pydap library can access the link:
    res = pydap.net.follow_redirect(test_url, session=session)
    assert(res.status_code == 200)

    # Check that the pydap library can access another link:
    res = pydap.net.follow_redirect(test_url_2, session=session)
    assert(res.status_code == 200)
    session.close()


@pytest.mark.auth
@pytest.mark.prod_url
@pytest.mark.skipif(not (os.environ.get('USERNAME_URS') and
                         os.environ.get('PASSWORD_URS')),
                    reason=('Without auth credentials, '
                            'this test cannot work'))
def test_basic_urs_query():
    session = urs.setup_session(os.environ.get('USERNAME_URS'),
                                os.environ.get('PASSWORD_URS'),
                                check_url=url)
    # Ensure authentication:
    res = pydap.net.follow_redirect(test_url, session=session)
    assert(res.status_code == 200)
    dataset = open_url(url, session=session)
    expected_data = [[[99066.15625, 99066.15625, 99066.15625,
                       99066.15625, 99066.15625],
                      [98868.15625, 98870.15625, 98872.15625,
                       98874.15625, 98874.15625],
                      [98798.15625, 98810.15625, 98820.15625,
                       98832.15625, 98844.15625],
                      [98856.15625, 98828.15625, 98756.15625,
                       98710.15625, 98776.15625],
                      [99070.15625, 99098.15625, 99048.15625,
                       98984.15625, 99032.15625]]]
    assert((dataset['SLP'][0, :5, :5] == expected_data).all())
    session.close()



"""pydap error response.

pydap will capture exceptions returned by the system and return them properly
formated as a DAP error response.

"""

from traceback import print_exception

from webob import Response
from six import StringIO, text_type

from ..lib import encode, __version__


class ErrorResponse(object):

    """A specialized response for errors.

    This is a special response used when an exception is captured and passed to
    the user as an Opendap error message:

    """

    def __init__(self, info):
        # get exception message
        buf = StringIO()
        print_exception(*info, file=buf)
        message = encode(buf.getvalue())

        # build error message
        code = getattr(info[0], 'code', -1)
        self.body = text_type('''Error {{
    code = {0};
    message = {1};
}}''').format(code, message)

    def __call__(self, environ, start_response):
        res = Response()
        res.text = self.body
        res.status = '500 Internal Error'
        res.content_type = 'text/plain'
        res.charset = 'utf-8'
        res.headers.add('Content-description', 'dods_error')
        res.headers.add('XDODS-Server', 'pydap/%s' % __version__)

        return res(environ, start_response)



"""Test the DODS response."""

import copy

import numpy as np
from webtest import TestApp as App
from webob.headers import ResponseHeaders
from collections import OrderedDict

from pydap.lib import START_OF_SEQUENCE, END_OF_SEQUENCE, __version__
from pydap.handlers.lib import BaseHandler
from pydap.tests.datasets import (
    VerySimpleSequence, SimpleSequence, SimpleGrid,
    SimpleArray, NestedSequence, SimpleStructure)
from pydap.responses.dods import dods, DODSResponse
import unittest


class TestDODSResponse(unittest.TestCase):

    """Test the DODS response."""

    def setUp(self):
        """Create a simple WSGI app."""
        app = App(BaseHandler(VerySimpleSequence))
        self.res = app.get('/.dods')

    def test_dispatcher(self):
        """Test that the dispatcher works on any object."""
        with self.assertRaises(StopIteration):
            dods(None)

    def test_status(self):
        """Test the status code."""
        self.assertEqual(self.res.status, "200 OK")

    def test_content_type(self):
        """Test the content type."""
        self.assertEqual(self.res.content_type, "application/octet-stream")

    def test_charset(self):
        """Test the charset."""
        self.assertEqual(self.res.charset, None)

    def test_headers(self):
        """Test the headers in the response."""
        self.assertEqual(
            self.res.headers,
            ResponseHeaders([
                ('XDODS-Server', 'pydap/' + __version__),
                ('Content-description', 'dods_data'),
                ('Content-type', 'application/octet-stream'),
                ('Access-Control-Allow-Origin', '*'),
                ('Access-Control-Allow-Headers',
                    'Origin, X-Requested-With, Content-Type'),
                ('Content-Length', '238')
            ]))

    def test_body(self):
        """Test response body."""
        dds, xdrdata = self.res.body.split(b'\nData:\n', 1)
        dds = dds.decode('ascii')
        self.assertEqual(dds, """Dataset {
    Sequence {
        Byte byte;
        Int32 int;
        Float32 float;
    } sequence;
} VerySimpleSequence;""")

        self.assertEqual(
            xdrdata,
            START_OF_SEQUENCE +
            b"\x00" b"\x00\x00\x00\x01" b"A \x00\x00" +
            START_OF_SEQUENCE +
            b"\x01" b"\x00\x00\x00\x02" b"A\xa0\x00\x00" +
            START_OF_SEQUENCE +
            b"\x02" b"\x00\x00\x00\x03" b"A\xf0\x00\x00" +
            START_OF_SEQUENCE +
            b"\x03" b"\x00\x00\x00\x04" b"B \x00\x00" +
            START_OF_SEQUENCE +
            b"\x04" b"\x00\x00\x00\x05" b"BH\x00\x00" +
            START_OF_SEQUENCE +
            b"\x05" b"\x00\x00\x00\x06" b"Bp\x00\x00" +
            START_OF_SEQUENCE +
            b"\x06" b"\x00\x00\x00\x07" b"B\x8c\x00\x00" +
            START_OF_SEQUENCE +
            b"\x07" b"\x00\x00\x00\x08" b"B\xa0\x00\x00" +
            END_OF_SEQUENCE)


class TestDODSResponseGrid(unittest.TestCase):

    """Test the DODS response with a grid."""

    def setUp(self):
        """Create a simple WSGI app."""
        app = App(BaseHandler(SimpleGrid))
        self.res = app.get('/.dods')

    def test_body(self):
        """Test the response body."""
        dds, xdrdata = self.res.body.split(b'\nData:\n', 1)
        dds = dds.decode('ascii')
        self.assertEqual(dds, """Dataset {
    Grid {
        Array:
            Int32 SimpleGrid[y = 2][x = 3];
        Maps:
            Int32 x[x = 3];
            Int32 y[y = 2];
    } SimpleGrid;
    Int32 x[x = 3];
    Int32 y[y = 2];
} SimpleGrid;""")

        self.assertEqual(
            xdrdata,
            b"\x00\x00\x00\x06"  # length
            b"\x00\x00\x00\x06"  # length, again
            b"\x00\x00\x00\x00"
            b"\x00\x00\x00\x01"
            b"\x00\x00\x00\x02"
            b"\x00\x00\x00\x03"
            b"\x00\x00\x00\x04"
            b"\x00\x00\x00\x05"

            b"\x00\x00\x00\x03"
            b"\x00\x00\x00\x03"
            b"\x00\x00\x00\x00"
            b"\x00\x00\x00\x01"
            b"\x00\x00\x00\x02"

            b"\x00\x00\x00\x02"
            b"\x00\x00\x00\x02"
            b"\x00\x00\x00\x00"
            b"\x00\x00\x00\x01"

            b"\x00\x00\x00\x03"
            b"\x00\x00\x00\x03"
            b"\x00\x00\x00\x00"
            b"\x00\x00\x00\x01"
            b"\x00\x00\x00\x02"

            b"\x00\x00\x00\x02"
            b"\x00\x00\x00\x02"
            b"\x00\x00\x00\x00"
            b"\x00\x00\x00\x01")


class TestDODSResponseStructure(unittest.TestCase):

    """Test the DODS response with a sequence that has a string."""

    def setUp(self):
        """Create a simple WSGI app."""
        app = App(BaseHandler(SimpleStructure))
        self.res = app.get('/.dods')

    def test_body(self):
        """Test response body."""
        dds, xdrdata = self.res.body.split(b'\nData:\n', 1)
        dds = dds.decode('ascii')
        self.assertEqual(dds, """Dataset {
    Structure {
        Int16 b;
        Byte ub;
        Int32 i32;
        UInt32 ui32;
        Int16 i16;
        UInt16 ui16;
        Float32 f32;
        Float64 f64;
        String s;
        String u;
        String U;
    } types;
} SimpleStructure;""")

        expected = OrderedDict([
            # -10 signed byte upconv. to Int32 for transfer
            ('b', (np.array(-10, dtype=np.byte)
                   .astype('>i')
                   .tobytes())),
            # 10 usigned byte padded to multiple of 4 bytes for transfer
            ('ub', (np.array(10, dtype=np.ubyte)
                    .tobytes()) + b'\x00\x00\x00'),
            ('i32', (np.array(-10, dtype=np.int32)
                     .astype('>i').tobytes())),
            ('ui32', (np.array(10, dtype=np.uint32)
                      .astype('>I').tobytes())),
            # -10 int16 upconv. to int32 for transfer
            ('i16', (np.array(-10, dtype=np.int16)
                     .astype('>i')
                     .newbyteorder('>').tobytes())),
            # 10 uint16 upconv. to uint32 for transfer
            ('ui16', (np.array(10, dtype=np.uint16)
                      .astype('>I')
                      .tobytes())),
            ('f32', (np.array(100, dtype=np.float32)
                     .astype('>f')
                     .tobytes())),
            ('f64', (np.array(1000, dtype=np.float64)
                     .astype('>d')
                     .tobytes())),
            ('s', b'\x00\x00\x00$This is a data test string (pass 0).\x00'),
            ('u', b'\x00\x00\x13http://www.dods.org\x00'),
            ('U', b'\x00\x00\x00\x0ctest unicode')])

        for key in expected:
            string = expected[key]
            assert xdrdata.startswith(string)
            xdrdata = xdrdata[len(string):]
        assert xdrdata == b''


class TestDODSResponseSequence(unittest.TestCase):

    """Test the DODS response with a sequence that has a string."""

    def setUp(self):
        """Create a simple WSGI app."""
        app = App(BaseHandler(SimpleSequence))
        self.res = app.get('/.dods?cast.id,cast.lon')

    def test_body(self):
        """Test response body."""
        dds, xdrdata = self.res.body.split(b'\nData:\n', 1)
        dds = dds.decode('ascii')
        self.assertEqual(dds, """Dataset {
    Sequence {
        String id;
        Int32 lon;
    } cast;
} SimpleSequence;""")

        self.assertEqual(
            xdrdata,
            START_OF_SEQUENCE +
            b"\x00\x00\x00\x01"   # length of the string (=1)
            b"1\x00\x00\x00"      # string zero-paddded to 4 bytes
            b"\x00\x00\x00d" +  # lon
            START_OF_SEQUENCE +
            b"\x00\x00\x00\x01"
            b"2\x00\x00\x00"
            b"\x00\x00\x00\xc8" +
            END_OF_SEQUENCE)


class TestDODSResponseArray(unittest.TestCase):

    """Test the DODS response with arrays of bytes and strings."""

    def setUp(self):
        """Create a simple WSGI app."""
        self.app = App(BaseHandler(SimpleArray))

    def test_body(self):
        """Test response body."""
        res = self.app.get("/.dods")
        dds, xdrdata = res.body.split(b'\nData:\n', 1)
        dds = dds.decode('ascii')
        self.assertEqual(dds, """Dataset {
    Byte byte[byte = 5];
    String string[string = 2];
    Int16 short;
} SimpleArray;""")

        self.assertEqual(
            xdrdata,
            b"\x00\x00\x00\x05"   # length of the byte array
            b"\x00\x00\x00\x05"   # length, again
            b"\x00"
            b"\x01"
            b"\x02"
            b"\x03"
            b"\x04"
            b"\x00\x00\x00"       # zero-padding to 4 bytes

            b"\x00\x00\x00\x02"   # length of the array, only once (string)
            b"\x00\x00\x00\x03"   # length of the first string
            b"one\x00"            # string padded to 4 bytes
            b"\x00\x00\x00\x03"
            b"two\x00"

            b"\x00\x00\x00\x01")  # shorts are encoded in 4 bytes

    def test_calculate_size_short(self):
        """Test that size is calculated correctly with shorts."""
        res = self.app.get("/.dods?short")
        self.assertEqual(res.headers["content-length"], "52")


class TestDODSResponseArrayterator(unittest.TestCase):

    """Test the DODS response when encountering an Arrayterator.

    In several cases the DODS response will work on an Arrayterator, which is
    a wrapper over arrays that allows iteration over them in blocks, instead of
    reading everything into memory.

    """

    def test_grid(self):
        """Test a grid."""
        modified = copy.copy(SimpleGrid)
        modified.SimpleGrid.SimpleGrid.data = np.lib.arrayterator.Arrayterator(
            modified.SimpleGrid.SimpleGrid.data, 2)

        app = App(BaseHandler(modified))
        res = app.get("/.dods?SimpleGrid.SimpleGrid")
        header_string = b"""Dataset {
    Structure {
        Int32 SimpleGrid[y = 2][x = 3];
    } SimpleGrid;
} SimpleGrid;
Data:
"""
        self.assertEqual(res.body, (header_string +
                                    b"\x00\x00\x00\x06"
                                    b"\x00\x00\x00\x06"
                                    b"\x00\x00\x00\x00"
                                    b"\x00\x00\x00\x01"
                                    b"\x00\x00\x00\x02"
                                    b"\x00\x00\x00\x03"
                                    b"\x00\x00\x00\x04"
                                    b"\x00\x00\x00\x05"))


class TestDODSResponseNestedSequence(unittest.TestCase):

    """Test the DODS response with nested sequences."""

    def test_iteration(self):
        """Test direct iteration over data."""
        response = DODSResponse(NestedSequence)
        output = b"".join(response)
        self.assertEqual(
            output, b"""Dataset {
    Sequence {
        Int32 lat;
        Int32 lon;
        Int32 elev;
        Sequence {
            Int32 time;
            Int32 slp;
            Int32 wind;
        } time_series;
    } location;
} NestedSequence;
Data:
""" +
            START_OF_SEQUENCE +
            b"\x00\x00\x00\x01"
            b"\x00\x00\x00\x01"
            b"\x00\x00\x00\x01" +
            START_OF_SEQUENCE +
                    b"\x00\x00\x00\n"
                    b"\x00\x00\x00\x0b"
                    b"\x00\x00\x00\x0c" +
            START_OF_SEQUENCE +
                    b"\x00\x00\x00\x15"
                    b"\x00\x00\x00\x16"
                    b"\x00\x00\x00\x17" +
            END_OF_SEQUENCE +
            START_OF_SEQUENCE +
                    b"\x00\x00\x00\x02"
                    b"\x00\x00\x00\x04"
                    b"\x00\x00\x00\x04" +
            START_OF_SEQUENCE +
                    b"\x00\x00\x00\x0f"
                    b"\x00\x00\x00\x10"
                    b"\x00\x00\x00\x11" +
            END_OF_SEQUENCE +
            START_OF_SEQUENCE +
                    b"\x00\x00\x00\x03"
                    b"\x00\x00\x00\x06"
                    b"\x00\x00\x00\t" +
            START_OF_SEQUENCE +
                    b"\x00\x00\x00\x04"
                    b"\x00\x00\x00\x08"
                    b"\x00\x00\x00\x10" +
                    START_OF_SEQUENCE +
                    b"\x00\x00\x00\x1f"
                    b"\x00\x00\x00 "
                    b"\x00\x00\x00!" +
                    START_OF_SEQUENCE +
                    b"\x00\x00\x00)"
                    b"\x00\x00\x00*"
                    b"\x00\x00\x00+" +
                    START_OF_SEQUENCE +
                    b"\x00\x00\x003"
                    b"\x00\x00\x004"
                    b"\x00\x00\x005" +
                    START_OF_SEQUENCE +
                    b"\x00\x00\x00="
                    b"\x00\x00\x00>"
                    b"\x00\x00\x00?" +
                    END_OF_SEQUENCE +
            END_OF_SEQUENCE)

    def test_body(self):
        """Test response body."""
        app = App(BaseHandler(NestedSequence))
        res = app.get("/.dods")
        dds, xdrdata = res.body.split(b'\nData:\n', 1)
        dds = dds.decode('ascii')
        self.assertEqual(dds, """Dataset {
    Sequence {
        Int32 lat;
        Int32 lon;
        Int32 elev;
        Sequence {
            Int32 time;
            Int32 slp;
            Int32 wind;
        } time_series;
    } location;
} NestedSequence;""")

        self.assertEqual(
            xdrdata,
            START_OF_SEQUENCE +
            b"\x00\x00\x00\x01"
            b"\x00\x00\x00\x01"
            b"\x00\x00\x00\x01" +
            START_OF_SEQUENCE +
            b"\x00\x00\x00\n"
            b"\x00\x00\x00\x0b"
            b"\x00\x00\x00\x0c" +
            START_OF_SEQUENCE +
            b"\x00\x00\x00\x15"
            b"\x00\x00\x00\x16"
            b"\x00\x00\x00\x17" +
            END_OF_SEQUENCE +
            START_OF_SEQUENCE +
            b"\x00\x00\x00\x02"
            b"\x00\x00\x00\x04"
            b"\x00\x00\x00\x04" +
            START_OF_SEQUENCE +
            b"\x00\x00\x00\x0f"
            b"\x00\x00\x00\x10"
            b"\x00\x00\x00\x11" +
            END_OF_SEQUENCE +
            START_OF_SEQUENCE +
            b"\x00\x00\x00\x03"
            b"\x00\x00\x00\x06"
            b"\x00\x00\x00\t" +
            START_OF_SEQUENCE +
            b"\x00\x00\x00\x04"
            b"\x00\x00\x00\x08"
            b"\x00\x00\x00\x10" +
            START_OF_SEQUENCE +
            b"\x00\x00\x00\x1f"
            b"\x00\x00\x00 "
            b"\x00\x00\x00!" +
            START_OF_SEQUENCE +
            b"\x00\x00\x00)"
            b"\x00\x00\x00*"
            b"\x00\x00\x00+" +
            START_OF_SEQUENCE +
            b"\x00\x00\x003"
            b"\x00\x00\x004"
            b"\x00\x00\x005" +
            START_OF_SEQUENCE +
            b"\x00\x00\x00="
            b"\x00\x00\x00>"
            b"\x00\x00\x00?" +
            END_OF_SEQUENCE +
            END_OF_SEQUENCE)



import multiprocessing
import time
import requests
import warnings

from werkzeug.serving import run_simple

from ..wsgi.ssf import ServerSideFunctions
from ..handlers.lib import BaseHandler
from .devel import DefaultDataset, LocalTestServer


def run_simple_server(application=BaseHandler(DefaultDataset),
                      port=8000, ssl_context=None):
    application = ServerSideFunctions(application)

    def app_check_for_shutdown(environ, start_response):
        if environ['PATH_INFO'].endswith('shutdown'):
            shutdown_server(environ)
            return shutdown_application(environ, start_response)
        else:
            return application(environ, start_response)

    run_simple('0.0.0.0', port,
               app_check_for_shutdown,
               ssl_context=ssl_context)


def shutdown_server(environ):
    if 'werkzeug.server.shutdown' not in environ:
        raise RuntimeError('Not running the development server')
    environ['werkzeug.server.shutdown']()


def shutdown_application(environ, start_response):
    start_response('200 OK', [('Content-Type', 'text/plain')])
    return [b'Server is shutting down.']


class LocalTestServerSSL(LocalTestServer):
    """
    Simple server instance that can be used to test pydap.
    Relies on multiprocessing and is usually slow (it has to
    start and shutdown which typically takes ~2 sec).

    Usage:
    >>> import numpy as np
    >>> from pydap.handlers.lib import BaseHandler
    >>> from pydap.model import DatasetType, BaseType
    >>> DefaultDataset = DatasetType("Default")
    >>> DefaultDataset["byte"] = BaseType("byte", np.arange(5, dtype="B"))
    >>> DefaultDataset["string"] = BaseType("string", np.array(["one", "two"]))
    >>> DefaultDataset["short"] = BaseType("short", np.array(1, dtype="h"))
    >>> DefaultDataset
    <DatasetType with children 'byte', 'string', 'short'>
    >>> application = BaseHandler(DefaultDataset)
    >>> from pydap.client import open_url

    As an instance:
    >>> with LocalTestServerSSL(application) as server:
    ...     dataset = open_url("http://localhost:%s" % server.port)
    ...     dataset
    ...     print(dataset['byte'].data[:])
    ...     print(dataset['string'].data[:])
    ...     print(dataset['short'].data[:])
    <DatasetType with children 'byte', 'string', 'short'>
    [0 1 2 3 4]
    [b'one' b'two']
    1

    Or by managing connection and deconnection:
    >>> server = LocalTestServerSSL(application)
    >>> server.start()
    >>> dataset = open_url("http://localhost:%s" % server.port)
    >>> dataset
    <DatasetType with children 'byte', 'string', 'short'>
    >>> print(dataset['byte'].data[:])
    [0 1 2 3 4]
    >>> server.shutdown()
    """
    def __init__(self, application=BaseHandler(DefaultDataset),
                 port=None, wait=0.5, polling=1e-2,
                 as_process=False, ssl_context=None):
        super(LocalTestServerSSL, self).__init__(application, port, wait,
                                                 polling, as_process)
        self._ssl_context = ssl_context

    @property
    def url(self):
        if self._ssl_context is None:
            return "http://{0}:{1}/".format(self._address, self.port)
        else:
            return "https://{0}:{1}/".format(self._address, self.port)

    def start(self):
        # Start a simple WSGI server:
        self._server = (multiprocessing
                        .Process(target=run_simple_server,
                                 args=(self.application,
                                       self.port,
                                       self._ssl_context)))
        self._server.start()
        # Wait a little while for the server to start:
        self.poll_server()

    def shutdown(self):
        # Shutdown the server:
        url = "http://0.0.0.0:%s/shutdown"
        if self._ssl_context is not None:
            url = url.replace('http', 'https')
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            requests.head(url % self.port, verify=False)
        time.sleep(self._wait)
        self._server.join()
        del(self._server)



from webtest import TestApp as App
import unittest


def MyApp(environ, start_response):
    start_response('200 OK', [('content-type', 'text/plain; charset=utf-8')])
    yield 'foo\n'.encode('utf-8')
    yield 'bar\n'.encode('utf-8')
    yield 'baz\n'.encode('utf-8')


class MyTest(unittest.TestCase):
    def setUp(self):
        self.app = App(MyApp)

    def test_webtest(self):
        expected = 'foo\nbar\nbaz\n'
        self.assertEqual(self.app.get('/').text, expected)



"""Test the pydap client."""

import os
import numpy as np
from pydap.handlers.lib import BaseHandler
from pydap.client import open_url, open_dods_url, open_file
from pydap.tests.datasets import SimpleSequence, SimpleGrid, SimpleStructure
import pytest


DODS = os.path.join(os.path.dirname(__file__), 'data/test.01.dods')
DAS = os.path.join(os.path.dirname(__file__), 'data/test.01.das')


@pytest.fixture
def sequence_app():
    return BaseHandler(SimpleSequence)


@pytest.fixture
def structure_app():
    return BaseHandler(SimpleStructure)


@pytest.mark.client
def test_open_url(sequence_app):
    """Open an URL and check dataset keys."""
    dataset = open_url('http://localhost:8001/', sequence_app)
    assert list(dataset.keys()) == ["cast"]


@pytest.mark.client
def test_open_dods():
    """Open a file downloaded from the test server with the DAS."""
    dataset = open_file(DODS)

    # test data
    assert dataset.data == [
        0, 1, 0, 0, 0, 0.0, 1000.0,
        'This is a data test string (pass 0).',
        'http://www.dods.org',
    ]

    # test attributes
    assert dataset.attributes == {}
    assert dataset.i32.attributes == {}
    assert dataset.b.attributes == {}


@pytest.mark.client
def test_open_dods_das():
    """Open a file downloaded from the test server with the DAS."""
    dataset = open_file(DODS, DAS)

    # test data
    assert dataset.data == [
        0, 1, 0, 0, 0, 0.0, 1000.0,
        'This is a data test string (pass 0).',
        'http://www.dods.org',
    ]

    # test attributes
    assert (dataset.i32.units == 'unknown')
    assert (dataset.i32.Description == 'A 32 bit test server int')
    assert (dataset.b.units == 'unknown')
    assert (dataset.b.Description == 'A test byte')
    assert (
        dataset.Facility['DataCenter'] ==
        "COAS Environmental Computer Facility")
    assert (
        dataset.Facility['PrincipleInvestigator'] ==
        ['Mark Abbott', 'Ph.D'])
    assert (
        dataset.Facility['DrifterType'] ==
        "MetOcean WOCE/OCM")


@pytest.mark.client
def test_open_dods_16bits(sequence_app):
    """Open the dods response from a server.

    Note that here we cannot simply compare ``dataset.data`` with the
    original data ``SimpleSequence.data``, because the original data
    contains int16 values which are transmitted as int32 in the DAP spec.

    """
    dataset = open_dods_url('.dods', application=sequence_app)
    assert (
        list(dataset.data) == [[
            ('1', 100, -10, 0, -1, 21, 35, 0),
            ('2', 200, 10, 500, 1, 15, 35, 100),
        ]])

    # attributes should be empty
    assert dataset.attributes == {}


@pytest.mark.client
def test_open_dods_with_attributes(sequence_app):
    """Open the dods response together with the das response."""
    dataset = open_dods_url('.dods', metadata=True, application=sequence_app)
    assert (dataset.NC_GLOBAL == {})
    assert (dataset.DODS_EXTRA == {})
    assert (
        dataset.description == "A simple sequence for testing.")
    assert (dataset.cast.lon.axis == 'X')
    assert (dataset.cast.lat.axis == 'Y')
    assert (dataset.cast.depth.axis == 'Z')
    assert (dataset.cast.time.axis == 'T')
    assert (
        dataset.cast.time.units == "days since 1970-01-01")


@pytest.fixture(scope='session')
def ssf_app():
    """Test the local implementation of server-side functions.

    Calling server-side functions is implemented using a lazy mechanism where
    arbitrary names are mapped to remove calls. The resulting dataset is only
    evaluated when ``__getitem__`` or ``__getattr__`` are called, allowing
    nested calls to be evaluated only once.

    """
    from pydap.wsgi.ssf import ServerSideFunctions
    return ServerSideFunctions(BaseHandler(SimpleGrid))


def test_original(ssf_app):
    """Test an unmodified call, without function calls."""
    original = open_url('/', application=ssf_app)
    assert (original.SimpleGrid.SimpleGrid.shape == (2, 3))


def test_first_axis(ssf_app):
    """Test mean over the first axis."""
    original = open_url('/', application=ssf_app)
    dataset = original.functions.mean(original.SimpleGrid, 0)
    assert (dataset.SimpleGrid.SimpleGrid.shape == (3,))
    np.testing.assert_array_equal(
        dataset.SimpleGrid.SimpleGrid.data,
        np.array([1.5, 2.5, 3.5]))


def test_second_axis(ssf_app):
    """Test mean over the second axis."""
    original = open_url('/', application=ssf_app)
    dataset = original.functions.mean(original.SimpleGrid, 1)
    assert (dataset.SimpleGrid.SimpleGrid.shape == (2,))
    np.testing.assert_array_equal(
        dataset.SimpleGrid.SimpleGrid.data,
        np.array([1.0, 4.0]))


def test_lazy_evaluation_getitem(ssf_app):
    """Test that the dataset is only loaded when accessed."""
    original = open_url('/', application=ssf_app)
    dataset = original.functions.mean(original.SimpleGrid, 0)
    assert dataset.dataset is None
    dataset['SimpleGrid']
    assert dataset.dataset is not None


def test_lazy_evaluation_getattr(ssf_app):
    """Test that the dataset is only loaded when accessed."""
    original = open_url('/', application=ssf_app)
    dataset = original.functions.mean(original.SimpleGrid, 0)
    assert dataset.dataset is None
    dataset.SimpleGrid
    assert dataset.dataset is not None


def test_nested_call(ssf_app):
    """Test nested calls."""
    original = open_url('/', application=ssf_app)
    dataset = original.functions.mean(
        original.functions.mean(original.SimpleGrid, 0), 0)
    assert (dataset['SimpleGrid']['SimpleGrid'].shape == ())
    np.testing.assert_array_equal(
        dataset.SimpleGrid.SimpleGrid.data,
        np.array(2.5))


def test_axis_mean(ssf_app):
    """Test the mean over an axis, returning a scalar."""
    original = open_url('/', application=ssf_app)
    dataset = original.functions.mean(original.SimpleGrid.x)
    assert (dataset.x.shape == ())
    np.testing.assert_array_equal(
        dataset.x.data,
        np.array(1.0))


@pytest.mark.client
def test_int16(structure_app):
    """Test that 16-bit values are represented correctly.

    Even though the DAP transfers int16 and uint16 as 32 bits, we want them to
    be represented locally using the correct type.

    Load an int16 -> should yield '>i2' type."""
    dataset = open_url("http://localhost:8001/", structure_app)
    assert (dataset.types.i16.dtype == np.dtype(">i2"))


@pytest.mark.client
def test_uint16(structure_app):
    """Test that 16-bit values are represented correctly.

    Even though the DAP transfers int16 and uint16 as 32 bits, we want them to
    be represented locally using the correct type.

    Load an uint16 -> should yield '>u2' type."""
    dataset = open_url("http://localhost:8001/", structure_app)
    assert (dataset.types.ui16.dtype == np.dtype(">u2"))



"""Test the DAP handler, which forms the core of the client."""

import numpy as np
from pydap.model import StructureType, GridType, DatasetType, BaseType
from pydap.handlers.lib import BaseHandler, ConstraintExpression
from pydap.handlers.dap import DAPHandler, BaseProxyDap2, SequenceProxy
from pydap.handlers.dap import find_pattern_in_string_iter
from pydap.tests.datasets import (
    SimpleSequence, SimpleGrid, SimpleArray, VerySimpleSequence)

import unittest
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


class TestDapHandler(unittest.TestCase):

    """Test that the handler creates the correct dataset from a URL."""

    def setUp(self):
        """Create WSGI apps"""
        self.app1 = BaseHandler(SimpleGrid)
        self.app2 = BaseHandler(SimpleSequence)
        self.app3 = BaseHandler(SimpleGrid, gzip=True)

    def test_grid(self):
        """Test that dataset has the correct data proxies for grids."""
        dataset = DAPHandler("http://localhost:8001/", self.app1).dataset

        self.assertEqual(list(dataset.keys()), ["SimpleGrid", "x", "y"])
        self.assertEqual(
            list(dataset.SimpleGrid.keys()), ["SimpleGrid", "x", "y"])

        # test one of the axis
        self.assertIsInstance(dataset.SimpleGrid.x.data, BaseProxyDap2)
        self.assertEqual(
            dataset.SimpleGrid.x.data.baseurl, "http://localhost:8001/")
        self.assertEqual(dataset.SimpleGrid.x.data.id, "SimpleGrid.x")
        self.assertEqual(dataset.SimpleGrid.x.data.dtype, np.dtype('>i4'))
        self.assertEqual(dataset.SimpleGrid.x.data.shape, (3,))
        self.assertEqual(
            dataset.SimpleGrid.x.data.slice, (slice(None),))

        # test the grid
        self.assertIsInstance(dataset.SimpleGrid.SimpleGrid.data, BaseProxyDap2)
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.baseurl,
            "http://localhost:8001/")
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.id, "SimpleGrid.SimpleGrid")
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.dtype, np.dtype('>i4'))
        self.assertEqual(dataset.SimpleGrid.SimpleGrid.data.shape, (2, 3))
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.slice,
            (slice(None), slice(None)))
        self.assertEqual(
                repr(dataset.SimpleGrid[:]),
                "<GridType with array 'SimpleGrid' and maps 'x', 'y'>")

    def test_grid_gzip(self):
        """Test that dataset has the correct data proxies for grids."""
        dataset = DAPHandler("http://localhost:8001/", self.app3).dataset

        self.assertEqual(list(dataset.keys()), ["SimpleGrid", "x", "y"])
        self.assertEqual(
            list(dataset.SimpleGrid.keys()), ["SimpleGrid", "x", "y"])

        # test one of the axis
        self.assertIsInstance(dataset.SimpleGrid.x.data, BaseProxyDap2)
        self.assertEqual(
            dataset.SimpleGrid.x.data.baseurl, "http://localhost:8001/")
        self.assertEqual(dataset.SimpleGrid.x.data.id, "SimpleGrid.x")
        self.assertEqual(dataset.SimpleGrid.x.data.dtype, np.dtype('>i4'))
        self.assertEqual(dataset.SimpleGrid.x.data.shape, (3,))
        self.assertEqual(
            dataset.SimpleGrid.x.data.slice, (slice(None),))

        # test the grid
        self.assertIsInstance(dataset.SimpleGrid.SimpleGrid.data, BaseProxyDap2)
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.baseurl,
            "http://localhost:8001/")
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.id, "SimpleGrid.SimpleGrid")
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.dtype, np.dtype('>i4'))
        self.assertEqual(dataset.SimpleGrid.SimpleGrid.data.shape, (2, 3))
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.slice,
            (slice(None), slice(None)))
        self.assertEqual(
                repr(dataset.SimpleGrid[:]),
                "<GridType with array 'SimpleGrid' and maps 'x', 'y'>")

    def test_grid_erddap(self):
        """Test that dataset has the correct data proxies for grids
           with the ERDDAP behavior."""
        with patch('pydap.handlers.lib.degenerate_grid_to_structure',
                   side_effect=(lambda x: x)) as mock_degenerate:
            dataset = DAPHandler("http://localhost:8001/", self.app1).dataset
            self.assertEqual(
                    repr(dataset.SimpleGrid[:]),
                    "<GridType with array 'SimpleGrid' and maps 'x', 'y'>")
            assert mock_degenerate.called

    def test_grid_output_grid_false(self):
        """Test that dataset has the correct data proxies for grids with
           option output_grid set to False."""
        dataset = DAPHandler("http://localhost:8001/", self.app1,
                             output_grid=False).dataset

        self.assertEqual(list(dataset.keys()), ["SimpleGrid", "x", "y"])
        self.assertEqual(
            list(dataset.SimpleGrid.keys()), ["SimpleGrid", "x", "y"])

        # test one of the axis
        self.assertIsInstance(dataset.SimpleGrid.x.data, BaseProxyDap2)
        self.assertEqual(
            dataset.SimpleGrid.x.data.baseurl, "http://localhost:8001/")
        self.assertEqual(dataset.SimpleGrid.x.data.id, "SimpleGrid.x")
        self.assertEqual(dataset.SimpleGrid.x.data.dtype, np.dtype('>i4'))
        self.assertEqual(dataset.SimpleGrid.x.data.shape, (3,))
        self.assertEqual(
            dataset.SimpleGrid.x.data.slice, (slice(None),))

        # test the grid
        self.assertIsInstance(dataset.SimpleGrid.SimpleGrid.data, BaseProxyDap2)
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.baseurl,
            "http://localhost:8001/")
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.id, "SimpleGrid.SimpleGrid")
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.dtype, np.dtype('>i4'))
        self.assertEqual(dataset.SimpleGrid.SimpleGrid.data.shape, (2, 3))
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.slice,
            (slice(None), slice(None)))
        np.testing.assert_array_equal(dataset.SimpleGrid[:],
                                      [[0, 1, 2], [3, 4, 5]])

    def test_grid_erddap_output_grid_false(self):
        """Test that dataset has the correct data proxies for grids with
           option output_grid set to False and with the ERDDAP behavior."""
        with patch('pydap.handlers.lib.degenerate_grid_to_structure',
                   side_effect=(lambda x: x)) as mock_degenerate:
            dataset = DAPHandler("http://localhost:8001/", self.app1,
                                 output_grid=False).dataset
            np.testing.assert_array_equal(dataset.SimpleGrid[:],
                                          [[0, 1, 2], [3, 4, 5]])
            assert mock_degenerate.called

    def test_grid_with_projection(self):
        """Test that a sliced proxy can be created for grids."""
        dataset = DAPHandler("http://localhost:8001/?SimpleGrid[0]",
                             self.app1).dataset

        self.assertEqual(dataset.SimpleGrid.x.data.shape, (1,))
        self.assertEqual(dataset.SimpleGrid.x.data.slice, (slice(0, 1, 1),))
        self.assertEqual(dataset.SimpleGrid.y.data.shape, (2,))
        self.assertEqual(dataset.SimpleGrid.y.data.slice, (slice(0, 3, 1),))
        self.assertEqual(dataset.SimpleGrid.SimpleGrid.data.shape, (1, 3))
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.slice,
            (slice(0, 1, 1), slice(0, 3, 1)))

    def test_base_type_with_projection(self):
        """Test that a sliced proxy can be created for a base type."""
        dataset = DAPHandler("http://localhost:8001/?x[1:1:2]",
                             self.app1).dataset

        self.assertEqual(dataset.x.data.shape, (2,))
        self.assertEqual(dataset.x.data.slice, (slice(1, 2, 1),))

    def test_grid_array_with_projection(self):
        """Test that a grid array can be properly pre sliced."""
        dataset = DAPHandler(
                             "http://localhost:8001/?SimpleGrid.SimpleGrid[0]",
                             self.app1).dataset

        # object should be a structure, not a grid
        self.assertEqual(list(dataset.keys()), ["SimpleGrid"])
        self.assertNotIsInstance(dataset.SimpleGrid, GridType)
        self.assertIsInstance(dataset.SimpleGrid, StructureType)

        self.assertEqual(dataset.SimpleGrid.SimpleGrid.data.shape, (1, 3))
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.slice,
            (slice(0, 1, 1), slice(0, 3, 1)))

    def test_grid_map_with_projection(self):
        """Test that a grid map can be properly pre sliced."""
        dataset = DAPHandler("http://localhost:8001/?SimpleGrid.x[0]",
                             self.app1).dataset

        self.assertEqual(dataset.SimpleGrid.x.data.shape, (1,))
        self.assertEqual(
            dataset.SimpleGrid.x.data.slice,
            (slice(0, 1, 1),))

    def test_sequence(self):
        """Test that dataset has the correct data proxies for sequences."""
        dataset = DAPHandler("http://localhost:8001/", self.app2).dataset

        self.assertEqual(list(dataset.keys()), ["cast"])
        self.assertEqual(
            list(dataset.cast.keys()), [
                'id', 'lon', 'lat', 'depth', 'time', 'temperature', 'salinity',
                'pressure'])

        # check the sequence
        self.assertIsInstance(dataset.cast.data, SequenceProxy)
        self.assertEqual(
            dataset.cast.data.baseurl, "http://localhost:8001/")
        self.assertEqual(dataset.cast.data.id, "cast")
        self.assertEqual(dataset.cast.data.shape, ())
        self.assertEqual(dataset.cast.data.selection, [])
        self.assertEqual(dataset.cast.data.slice, (slice(None),))

        # check a child
        self.assertIsInstance(dataset.cast.lon.data, SequenceProxy)
        self.assertEqual(
            dataset.cast.lon.data.baseurl, "http://localhost:8001/")
        self.assertEqual(dataset.cast.lon.data.id, "cast.lon")
        self.assertEqual(dataset.cast.lon.data.shape, ())
        self.assertEqual(dataset.cast.lon.data.dtype, np.dtype(">i4"))
        self.assertEqual(dataset.cast.lon.data.selection, [])
        self.assertEqual(dataset.cast.lon.data.slice, (slice(None),))

    def test_sequence_with_projection(self):
        """Test projections applied to sequences."""
        dataset = DAPHandler(
            "http://localhost:8001/?cast[1]", self.app2).dataset

        self.assertEqual(dataset.cast.data.slice, (slice(1, 2, 1),))
        self.assertEqual(
            [tuple(row) for row in dataset.cast.iterdata()], [
                ('2', 200, 10, 500, 1, 15, 35, 100)])


class TestBaseProxy(unittest.TestCase):

    """Test `BaseProxy` objects."""

    def setUp(self):
        """Create a WSGI app"""
        self.app = BaseHandler(SimpleArray)

        self.data = BaseProxyDap2(
                              "http://localhost:8001/", "byte",
                              np.dtype("B"), (5,),
                              application=self.app)

    def test_repr(self):
        """Test the object representation."""
        self.assertEqual(
                         repr(self.data),
                         "BaseProxy('http://localhost:8001/', 'byte', "
                         "dtype('uint8'), (5,), "
                         "(slice(None, None, None),))")

    def test_getitem(self):
        """Test the ``__getitem__`` method."""
        np.testing.assert_array_equal(self.data[:], np.arange(5))

    def test_getitem_out_of_bound(self):
        """
        Test the ``__getitem__`` method with out of bounds
        slices.
        The same result as numpy is expected.
        """
        np.testing.assert_array_equal(self.data[:10], np.arange(5)[:10])

    def test_inexact_division(self):
        """Test data retrieval when step > 1 and division inexact."""
        np.testing.assert_array_equal(self.data[0:3:2], np.arange(5)[0:3:2])

    def test_len(self):
        """Test length method."""
        self.assertEqual(len(self.data), 5)

    def test_iteration(self):
        """Test iteration."""
        self.assertEqual(list(iter(self.data)), list(range(5)))

    def test_comparisons(self):
        """Test all the comparisons."""
        np.testing.assert_array_equal(self.data == 2, np.arange(5) == 2)
        np.testing.assert_array_equal(self.data != 2, np.arange(5) != 2)
        np.testing.assert_array_equal(self.data >= 2, np.arange(5) >= 2)
        np.testing.assert_array_equal(self.data <= 2, np.arange(5) <= 2)
        np.testing.assert_array_equal(self.data > 2, np.arange(5) > 2)
        np.testing.assert_array_equal(self.data < 2, np.arange(5) < 2)


class TestBaseProxyShort(unittest.TestCase):

    """Test `BaseProxy` objects with short dtype."""

    def setUp(self):
        """Create a WSGI app with array data"""
        self.app = BaseHandler(SimpleArray)

        self.data = BaseProxyDap2(
                              "http://localhost:8001/", "short",
                              np.dtype(">h"), (),
                              application=self.app)

    def test_getitem(self):
        """Test the ``__getitem__`` method."""
        np.testing.assert_array_equal(self.data[:], np.array(1))
        assert self.data[:].dtype.char == 'h'


class TestBaseProxyString(unittest.TestCase):

    """Test a ``BaseProxy`` with string data."""

    def setUp(self):
        """Create a WSGI app with array data"""
        dataset = DatasetType("test")
        dataset["s"] = BaseType("s", np.array(["one", "two", "three"]))
        self.app = BaseHandler(dataset)

        self.data = BaseProxyDap2(
                              "http://localhost:8001/", "s",
                              np.dtype("|S5"), (3,), application=self.app)

    def test_getitem(self):
        """Test the ``__getitem__`` method."""
        np.testing.assert_array_equal(
            self.data[:], np.array(["one", "two", "three"], dtype='S'))


class TestSequenceProxy(unittest.TestCase):

    """Test that a ``SequenceProxy`` behaves like a Numpy structured array."""

    def setUp(self):
        """Create a WSGI app"""
        self.app = BaseHandler(VerySimpleSequence)
        self.local = VerySimpleSequence.sequence.data

        dataset = DAPHandler("http://localhost:8001/", self.app).dataset
        self.remote = dataset.sequence.data

    def test_repr(self):
        """Test the object representation."""
        self.assertEqual(
            repr(self.remote),
            "SequenceProxy('http://localhost:8001/', "
            "<SequenceType with children 'byte', 'int', 'float'>, [], "
            "(slice(None, None, None),))")

    def test_attributes(self):
        """Test attributes of the remote sequence."""
        self.assertEqual(self.remote.baseurl, "http://localhost:8001/")
        self.assertEqual(self.remote.id, "sequence")
        self.assertEqual(list(self.remote.template.keys()),
                         ["byte", "int", "float"])
        self.assertEqual(self.remote.selection, [])
        self.assertEqual(self.remote.slice, (slice(None),))

    def test_getitem(self):
        """Test modifications to the proxy object."""
        child = self.remote["int"]
        self.assertEqual(child.id, "sequence.int")
        self.assertEqual(child.template.dtype, np.dtype(">i4"))
        self.assertEqual(child.template.shape, ())

        print(self.remote)
        child = self.remote[["float", "int"]]
        print(child)
        self.assertEqual(child.id, "sequence.float,sequence.int")
        self.assertEqual(list(child.template.keys()), ["float", "int"])

        child = self.remote[0]
        self.assertEqual(child.slice, (slice(0, 1, 1),))

        child = self.remote[0:2]
        self.assertEqual(child.slice, (slice(0, 2, 1),))

    def test_url(self):
        """Test URL generation."""
        self.assertEqual(
            self.remote.url, "http://localhost:8001/.dods?sequence")
        self.assertEqual(
            self.remote["int"].url, "http://localhost:8001/.dods?sequence.int")
        self.assertEqual(
            self.remote[["float", "int"]].url,
            "http://localhost:8001/.dods?sequence.float,sequence.int")

    def test_iter(self):
        """Test iteration."""
        self.assertEqual(
            [tuple(row) for row in self.local], [
                (0, 1, 10.0),
                (1, 2, 20.0),
                (2, 3, 30.0),
                (3, 4, 40.0),
                (4, 5, 50.0),
                (5, 6, 60.0),
                (6, 7, 70.0),
                (7, 8, 80.0)])

        self.assertEqual(
            [tuple(row) for row in self.remote], [
                (0, 1, 10.0),
                (1, 2, 20.0),
                (2, 3, 30.0),
                (3, 4, 40.0),
                (4, 5, 50.0),
                (5, 6, 60.0),
                (6, 7, 70.0),
                (7, 8, 80.0)])

    def test_iter_child(self):
        """Test iteration over a child."""
        child = self.local["byte"]
        self.assertEqual(list(child), [0, 1, 2, 3, 4, 5, 6, 7])

        child = self.remote["byte"]
        self.assertEqual(list(child), [0, 1, 2, 3, 4, 5, 6, 7])

    def test_iter_find_pattern(self):
        pattern = b'Data:\n'
        # Check in a simple iteration:
        string_iter = iter([b'blahblah ', b'Data:\n', b'blahblahblah'])
        last_chunk = find_pattern_in_string_iter(pattern, string_iter)
        assert last_chunk == b''
        assert next(string_iter) == b'blahblahblah'

        # Check in a more complex iteration:
        string_iter = iter([b'blahblah Da', b'ta:\nblahb', b'lahblah'])
        last_chunk = find_pattern_in_string_iter(pattern, string_iter)
        assert last_chunk == b'blahb'
        assert next(string_iter) == b'lahblah'

        # Check for a character by character iteration:
        string_iter = iter([b'b', b'l', b'a', b'h', b'b', b'l', b'a',
                            b'h', b' ', b'D', b'a', b't', b'a', b':', b'\n',
                            b'b', b'l', b'a'])
        last_chunk = find_pattern_in_string_iter(pattern, string_iter)
        assert last_chunk == b''
        assert next(string_iter) == b'b'

    def test_comparisons(self):
        """Test lazy comparisons on the object."""
        filtered = self.remote[self.remote["byte"] == 4]
        self.assertEqual(filtered.selection, ["sequence.byte=4"])

        filtered = self.remote[self.remote["byte"] != 4]
        self.assertEqual(filtered.selection, ["sequence.byte!=4"])

        filtered = self.remote[self.remote["byte"] >= 4]
        self.assertEqual(filtered.selection, ["sequence.byte>=4"])

        filtered = self.remote[self.remote["byte"] <= 4]
        self.assertEqual(filtered.selection, ["sequence.byte<=4"])


class TestSequenceWithString(unittest.TestCase):

    def setUp(self):
        """Create a WSGI app"""
        self.app = BaseHandler(SimpleSequence)
        self.local = SimpleSequence.cast.data

        dataset = DAPHandler("http://localhost:8001/", self.app).dataset
        self.remote = dataset.cast.data

    def test_iter(self):
        """Test iteration."""
        self.assertEqual(
            [tuple(row) for row in self.local], [
                ('1', 100, -10, 0, -1, 21, 35, 0),
                ('2', 200, 10, 500, 1, 15, 35, 100)])

        self.assertEqual(
            [tuple(row) for row in self.remote], [
                ('1', 100, -10, 0, -1, 21, 35, 0),
                ('2', 200, 10, 500, 1, 15, 35, 100)])

    def test_projection(self):
        """Test if we can select only a few variables."""
        filtered = self.local[["salinity", "depth"]]
        self.assertEqual(
            [tuple(row) for row in filtered], [(35, 0), (35, 500)])

        filtered = self.remote[["salinity", "depth"]]
        self.assertEqual(
            [tuple(row) for row in filtered], [(35, 0), (35, 500)])

    def test_iter_child(self):
        """Test iteration over a child."""
        child = self.local["salinity"]
        self.assertEqual(list(child), [35, 35])

        child = self.remote["salinity"]
        self.assertEqual(list(child), [35, 35])

    def test_filtering(self):
        """Test filtering of data."""
        filtered = self.local[self.local["lon"] > 100]
        self.assertEqual(
            [tuple(row) for row in filtered], [
                ('2', 200, 10, 500, 1, 15, 35, 100)])

        filtered = self.remote[self.remote["lon"] > 100]
        self.assertEqual(
            [tuple(row) for row in filtered], [
                ('2', 200, 10, 500, 1, 15, 35, 100)])

        # filtering works because we store comparisons as lazy objects
        self.assertIsInstance(self.remote["lon"] > 100, ConstraintExpression)
        self.assertEqual(filtered.selection, ['cast.lon>100'])


class TestStringBaseType(unittest.TestCase):

    """Regression test for string base type."""

    def setUp(self):
        """Create a WSGI app with array data"""
        dataset = DatasetType("test")
        data = np.array("This is a test", dtype='S')
        dataset["s"] = BaseType("s", data)
        self.app = BaseHandler(dataset)

        self.data = BaseProxyDap2(
                              "http://localhost:8001/", "s",
                              np.dtype("|S14"), (), application=self.app)

    def test_getitem(self):
        """Test the ``__getitem__`` method."""
        np.testing.assert_array_equal(self.data[:],
                                      "This is a test")


class TestArrayStringBaseType(unittest.TestCase):

    """Regression test for an array of unicode base type."""

    def setUp(self):
        """Create a WSGI app with array data"""
        dataset = DatasetType("test")
        self.original_data = np.array([["This ", "is "], ["a ", "test"]],
                                      dtype='<U5')
        dataset["s"] = BaseType("s", self.original_data)
        self.app = BaseHandler(dataset)

        self.data = DAPHandler("http://localhost:8001/",  self.app).dataset.s

    def test_getitem(self):
        """Test the ``__getitem__`` method."""
        np.testing.assert_array_equal(self.data[:], self.original_data)



"""Test DAP4 parsing functions."""

import pydap.parsers.dmr
import os


def load_dmr_file(file_path):
    abs_path = os.path.join(os.path.dirname(__file__), file_path)
    with open(abs_path, 'r') as dmr_file:
        dmr = dmr_file.read()
    dataset = pydap.parsers.dmr.dmr_to_dataset(dmr)
    return dataset


def test_single_scalar():
    """Test a single scalar case."""
    single_scalar_dmr = """<Dataset name="foo"><Int32 name="x"/></Dataset>"""
    dataset = pydap.parsers.dmr.dmr_to_dataset(single_scalar_dmr)
    assert dataset['x'].dtype == '>i4'


def test_coads_climatology2():
    dataset = load_dmr_file('data/dmrs/coads_climatology.nc.dmr')
    assert dataset['SST'].attributes['long_name'] == 'SEA SURFACE TEMPERATURE'
    assert dataset['SST'].attributes['missing_value'] == '-9.99999979e+33'
    assert dataset['AIRT'].shape == (12, 90, 180)
    assert dataset['VWND'].dtype.str == '>f4'


def test_atl03():
    # Contains groups
    dataset = load_dmr_file('data/dmrs/ATL03_20181228015957_13810110_003_01.2var.h5.dmrpp.dmr')
    assert dataset['/gt1r/bckgrd_atlas/bckgrd_int_height'].attributes['contentType'] == 'modelResult'


def test_jpl():
    # Contains groups
    dataset = load_dmr_file('data/dmrs/20220102090000-JPL-L4_GHRSST-SSTfnd-MUR-GLOB-v02.0-fv04.1.dmr')
    assert dataset['sea_ice_fraction'].dimensions == ['time', 'lat', 'lon']


def test_mod05():
    dataset = load_dmr_file('data/dmrs/MOD05_L2.A2019336.2315.061.2019337071952.hdf.dmr')
    assert dataset['Water_Vapor_Infrared'].dimensions == ['Cell_Along_Swath_5km', 'Cell_Across_Swath_5km']








import io
import os
from pathlib import Path
import re
from setuptools import setup, find_packages
import sys

install_requires = [
    'numpy',
    'Webob',
    'Jinja2',
    'docopt',
    'six >= 1.4.0',
    'beautifulsoup4',
    'requests'
]

if sys.version_info < (3, 5):
    install_requires.append('singledispatch')

functions_extras = [
    'gsw==3.0.6',
    'coards'
]

server_extras = [
    'gunicorn',
    'PasteDeploy',
]

docs_extras = [
    'Sphinx',
    'Pygments',
    'sphinx_rtd_theme',
]

cas_extras = [
    'lxml'
]

hdl_netcdf_extras = [
    'netCDF4',
    'ordereddict'
]

# These extras should only contain pure
# testing packages (e.g. pytest, mock, flake8, ...)
tests_extras = [
    'pytest>=3.6',
    'pytest-cov',
    'pytest-attrib',
    'beautifulsoup4',
    'requests-mock',
    'WebTest',
    'flake8',
    'werkzeug',
    'pyopenssl']


testing_extras = (functions_extras + cas_extras + server_extras +
                  hdl_netcdf_extras + tests_extras)

if sys.version_info < (3, 3):
    testing_extras.append('mock')


def read(filename, encoding='utf-8'):
    """read file contents"""
    full_path = os.path.join(os.path.dirname(__file__), filename)
    with io.open(full_path, encoding=encoding) as fh:
        contents = fh.read().strip()
    return contents


def get_package_version():
    """get version from top-level package init"""
    version_file = read('src/pydap/__init__.py')
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


def get_long_description():
    """ Return contents of README.md as the long_description. """
    return (Path(__file__).parent / "README.md" ).read_text()


setup(name='pydap',
      version=get_package_version(),
      description="An implementation of the Data Access Protocol.",
      long_description=get_long_description(),
      long_description_content_type='text/markdown',
      classifiers=[
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11"
            # Get strings from
            # http://pypi.python.org/pypi?%3Aaction=list_classifiers
            ],
      keywords='opendap dods dap dap2 dap4 science data',
      author='Roberto De Almeida',
      author_email='roberto@dealmeida.net',
      maintainer='James Hiebert',
      maintainer_email='james@hiebert.name',
      url='http://pydap.org/',
      license='MIT',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      include_package_data=True,
      zip_safe=False,
      namespace_packages=["pydap", "pydap.responses",
                          "pydap.handlers", "pydap.tests"],
      install_requires=install_requires,
      extras_require={
            'client': [],
            'tests': tests_extras,
            'docs': docs_extras,
            'functions': functions_extras,
            'cas': cas_extras,
            'handlers.netcdf': hdl_netcdf_extras,
            'netcdf': hdl_netcdf_extras,
            'server': server_extras,
            'handlers.netcdf': hdl_netcdf_extras,
            'testing': testing_extras
      },
      test_suite="pydap.tests",
      entry_points="""
            [pydap.handler]
            nc = pydap.handlers.netcdf:NetCDFHandler

            [pydap.response]
            das = pydap.responses.das:DASResponse
            dds = pydap.responses.dds:DDSResponse
            dods = pydap.responses.dods:DODSResponse
            html = pydap.responses.html:HTMLResponse
            asc = pydap.responses.ascii:ASCIIResponse
            ascii = pydap.responses.ascii:ASCIIResponse
            ver = pydap.responses.version:VersionResponse

            [pydap.function]
            bounds = pydap.wsgi.functions:bounds
            mean = pydap.wsgi.functions:mean
            density = pydap.wsgi.functions:density

            [console_scripts]
            pydap = pydap.wsgi.app:main
            dods = pydap.handlers.dap:dump
      """, python_requires='>=3.7')



from bs4 import BeautifulSoup
from six.moves.urllib.parse import urlsplit, urlunsplit
import warnings
import requests
from requests.packages.urllib3.exceptions import (InsecureRequestWarning,
                                                  InsecurePlatformWarning)
import copy
from .. import lib

ssl_verify_categories = [InsecureRequestWarning,
                         InsecurePlatformWarning]


def setup_session(uri,
                  username=None,
                  password=None,
                  check_url=None,
                  session=None,
                  verify=True,
                  username_field='username',
                  password_field='password'):
    '''
    A general function to set-up requests session with cookies
    using beautifulsoup and by calling the right url.
    '''

    if session is None:
        # Connections must be closed since some CAS
        # will cough when connections are kept alive:
        headers = [('User-agent', 'pydap/{}'.format(lib.__version__)),
                   ('Connection', 'close')]
        session = requests.Session()
        session.headers.update(headers)

    if uri is None:
        return session

    if not verify:
        verify_flag = session.verify
        session.verify = False

    if isinstance(uri, str):
        url = uri
    else:
        url = uri(check_url)

    if password is None or password == '':
        warnings.warn('password was not set. '
                      'this was likely unintentional '
                      'but will result is much fewer datasets.')
        if not verify:
            session.verify = verify_flag
        return session

    # Allow for several subsequent security layers:
    full_url = copy.copy(url)
    if isinstance(full_url, list):
        url = full_url[0]

    with warnings.catch_warnings():
        if not verify:
            # Catch warnings. It is assumed that the
            # user that explicitly uses verify=False
            # is either fully aware of the risks
            # or cannot avoid the risks because of
            # an improperly configured server.
            # This error will usually occur with
            # ESGF authentication.
            for category in ssl_verify_categories:
                warnings.filterwarnings("ignore",
                                        category=category)

        response = soup_login(session, url, username, password,
                              username_field=username_field,
                              password_field=password_field)

        # If there are further security levels.
        # At the moment only used for CEDA OPENID:
        if (isinstance(full_url, list) and
           len(full_url) > 1):
            for url in full_url[1:]:
                response = soup_login(session, response.url,
                                      username, password,
                                      username_field=None,
                                      password_field=None)
        response.close()

        if check_url:
            if (username is not None and
               password is not None):
                res = session.get(check_url, auth=(username, password))
                if res.status_code == 401:
                    res = session.get(res.url, auth=(username, password))
                res.close()
            raise_if_form_exists(check_url, session)

    if not verify:
        session.verify = verify_flag
    return session


def raise_if_form_exists(url, session):
    """
    This function raises a UserWarning if the link has forms
    """

    user_warning = ('Navigate to {0}, '.format(url) +
                    'login and follow instructions. '
                    'It is likely that you have to perform some one-time '
                    'registration steps before acessing this data.')

    resp = session.get(url)
    soup = BeautifulSoup(resp.content, 'lxml')
    if len(soup.select('form')) > 0:
        raise UserWarning(user_warning)


def soup_login(session, url, username, password,
               username_field='username',
               password_field='password'):
    resp = session.get(url)

    soup = BeautifulSoup(resp.content, 'lxml')
    login_form = soup.select('form')[0]

    def get_to_url(current_url, to_url):
        split_current = urlsplit(current_url)
        split_to = urlsplit(to_url)
        comb = [val2 if val1 == '' else val1
                for val1, val2 in zip(split_to, split_current)]
        return urlunsplit(comb)
    to_url = get_to_url(resp.url, login_form.get('action'))

    session.headers['Referer'] = resp.url

    payload = {}
    if username_field is not None:
        if len(login_form.findAll('input', {'name': username_field})) > 0:
            payload.update({username_field: username})

    if password_field is not None:
        if len(login_form.findAll('input', {'name': password_field})) > 0:
            payload.update({password_field: password})
        else:
            # If there is no password_field, it might be because
            # something should be handled in the browser
            # for the first attempt. This is common when using
            # pydap with the ESGF for the first time.
            raise Exception('Navigate to {0}. '
                            'If you are unable to '
                            'login, you must either '
                            'wait or use authentication '
                            'from another service.'
                            .format(url))

    # Replicate all other fields:
    for input in login_form.findAll('input'):
        if (input.get('name') not in payload and
           input.get('name') is not None):
            payload.update({input.get('name'): input.get('value')})

    # Remove other submit fields:
    submit_type = 'submit'
    submit_names = [input.get('name') for input
                    in login_form.findAll('input', {'type': submit_type})]
    for input in login_form.findAll('input', {'type': submit_type}):
        if ('submit' in submit_names and
           input.get('name').lower() != 'submit'):
            payload.pop(input.get('name'), None)

    return session.post(to_url, data=payload)
